import keras

from bayesflow.types import Tensor
from bayesflow.utils import (
    find_permutation,
    layer_kwargs,
    weighted_mean,
)
from bayesflow.utils.serialization import deserialize, serializable, serialize

from .actnorm import ActNorm
from .couplings import DualCoupling
from ..inference_network import InferenceNetwork


@serializable
class CouplingFlow(InferenceNetwork):
    """Implements a coupling flow as a sequence of dual couplings with permutations and activation
    normalization. Incorporates ideas from [1-5].

    [1] Kingma, D. P., & Dhariwal, P. (2018).
    Glow: Generative flow with invertible 1x1 convolutions.
    Advances in Neural Information Processing Systems, 31.

    [2] Durkan, C., Bekasov, A., Murray, I., & Papamakarios, G. (2019).
    Neural spline flows. Advances in Neural Information Processing Systems, 32.

    [3] Ardizzone, L., Kruse, J., Lüth, C., Bracher, N., Rother, C., & Köthe, U. (2020).
    Conditional invertible neural networks for diverse image-to-image translation.
    In DAGM German Conference on Pattern Recognition (pp. 373-387). Springer, Cham.

    [4] Radev, S. T., Mertens, U. K., Voss, A., Ardizzone, L., & Köthe, U. (2020).
    BayesFlow: Learning complex stochastic simulators with invertible neural networks.
    IEEE Transactions on Neural Networks and Learning Systems.

    [5] Alexanderson, S., & Henter, G. E. (2020).
    Robust model training and generalisation with Studentising flows.
    arXiv preprint arXiv:2006.06599.
    """

    def __init__(
        self,
        subnet: str | type = "mlp",
        depth: int = 6,
        transform: str = "affine",
        permutation: str | None = "random",
        use_actnorm: bool = True,
        base_distribution: str = "normal",
        subnet_kwargs: dict[str, any] = None,
        transform_kwargs: dict[str, any] = None,
        **kwargs,
    ):
        """
        Initializes an invertible flow-based model with a sequence of transformations.

        This model constructs a deep invertible architecture composed of multiple
        layers, including ActNorm, learned permutations, and coupling layers.

        The specific transformation applied in the coupling layers is determined by
        `transform`, while the subnet type can be either an MLP or another callable
        architecture specified by `subnet`. If `use_actnorm` is set to True, an
        ActNorm layer is applied before each coupling layer.

        The model can be initialized with a base distribution, such as a standard normal, for
        density estimation. It can also use more flexible distributions, e.g., GMMs for
        highly multimodal, low-dimensional distributions or Multivariate Student-t for
        heavy-tailed distributions.

        Parameters
        ----------
        subnet : str or type, optional
            The architecture type used within the coupling layers. Can be a string
            identifier like "mlp" or a callable type. Default is "mlp".
        depth : int, optional
            The number of invertible layers in the model. Default is 6.
        transform : str, optional
            The type of transformation used in the coupling layers, such as "affine".
            Default is "affine".
        permutation : str or None, optional
            The type of permutation applied between layers. Can be "orthogonal", "random", "swap", or None
            (no permutation). Default is "random".
        use_actnorm : bool, optional
            Whether to apply ActNorm before each coupling layer. Default is True.
        base_distribution : str, optional
            The base probability distribution from which samples are drawn, such as
            "normal". Default is "normal".
        subnet_kwargs : dict of str to any, optional
            Keyword arguments forwarded to the subnet (e.g., MLP) constructor within
            each coupling layer, such as hidden sizes or activation choices.
        transform_kwargs : dict of str to any, optional
            Keyword arguments forwarded to the affine or spline transforms
            (e.g., bins for splines)
        **kwargs
            Additional keyword arguments passed to `InvertibleLayer`.

        """
        super().__init__(base_distribution=base_distribution, **kwargs)

        self.subnet = subnet
        self.depth = depth
        self.transform = transform
        self.permutation = permutation
        self.use_actnorm = use_actnorm

        self.invertible_layers = []
        for i in range(depth):
            if use_actnorm:
                self.invertible_layers.append(ActNorm())

            if (p := find_permutation(permutation)) is not None:
                self.invertible_layers.append(p)

            self.invertible_layers.append(
                DualCoupling(subnet, transform, subnet_kwargs=subnet_kwargs, transform_kwargs=transform_kwargs)
            )

        # We only need to do this from coupling flows, since we do not serialize invertible layers
        self.subnet_kwargs = subnet_kwargs
        self.transform_kwargs = transform_kwargs

    # noinspection PyMethodOverriding
    def build(self, xz_shape, conditions_shape=None):
        for layer in self.invertible_layers:
            layer.build(xz_shape=xz_shape, conditions_shape=conditions_shape)

        self.base_distribution.build(xz_shape)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()
        base_config = layer_kwargs(base_config)

        config = {
            "subnet": self.subnet,
            "depth": self.depth,
            "transform": self.transform,
            "permutation": self.permutation,
            "use_actnorm": self.use_actnorm,
            "base_distribution": self.base_distribution,
            "subnet_kwargs": self.subnet_kwargs,
            "transform_kwargs": self.transform_kwargs,
        }

        return base_config | serialize(config)

    def _forward(
        self, x: Tensor, conditions: Tensor = None, density: bool = False, training: bool = False, **kwargs
    ) -> Tensor | tuple[Tensor, Tensor]:
        z = x
        log_det = keras.ops.zeros(keras.ops.shape(x)[:-1])
        for layer in self.invertible_layers:
            z, det = layer(z, conditions=conditions, inverse=False, training=training)
            log_det += det

        if density:
            log_density_latent = self.base_distribution.log_prob(z)
            log_density = log_density_latent + log_det
            return z, log_density

        return z

    def _inverse(
        self, z: Tensor, conditions: Tensor = None, density: bool = False, training: bool = False, **kwargs
    ) -> Tensor | tuple[Tensor, Tensor]:
        x = z
        log_det = keras.ops.zeros(keras.ops.shape(z)[:-1])
        for layer in reversed(self.invertible_layers):
            x, det = layer(x, conditions=conditions, inverse=True, training=training)
            log_det += det

        if density:
            log_prob = self.base_distribution.log_prob(z)
            log_density = log_prob - log_det
            return x, log_density

        return x

    def compute_metrics(
        self, x: Tensor, conditions: Tensor = None, sample_weight: Tensor = None, stage: str = "training"
    ) -> dict[str, Tensor]:
        base_metrics = super().compute_metrics(x, conditions=conditions, stage=stage)

        z, log_density = self(x, conditions=conditions, inverse=False, density=True)
        loss = weighted_mean(-log_density, sample_weight)

        return base_metrics | {"loss": loss}



import keras

from bayesflow.types import Tensor
from bayesflow.utils import find_recurrent_net, layer_kwargs
from bayesflow.utils.decorators import sanitize_input_shape
from bayesflow.utils.serialization import serializable


@serializable
class SkipRecurrentNet(keras.Layer):
    """
    Implements a Skip recurrent layer as described in [1], allowing a more flexible recurrent backbone
    and a more efficient implementation.

    [1] Y. Zhang and L. Mikelsons, Solving Stochastic Inverse Problems with Stochastic BayesFlow,
    2023 IEEE/ASME International Conference on Advanced Intelligent Mechatronics (AIM),
    Seattle, WA, USA, 2023, pp. 966-972, doi: 10.1109/AIM46323.2023.10196190.
    """

    def __init__(
        self,
        hidden_dim: int = 256,
        recurrent_type: str = "gru",
        bidirectional: bool = True,
        input_channels: int = 64,
        skip_steps: int = 4,
        dropout: float = 0.05,
        **kwargs,
    ):
        """
        Creates a skip recurrent neural network layer that extends a traditional recurrent backbone with
        skip connections implemented via convolution and an additional recurrent path. This allows
        more efficient modeling of long-term dependencies by combining local and non-local temporal
        features.

        Parameters
        ----------
        hidden_dim : int, optional
            Dimensionality of the hidden state in the recurrent layers. Default is 256.
        recurrent_type : str, optional
            Type of recurrent unit to use. Should correspond to a supported type in `find_recurrent_net`,
            such as "gru" or "lstm". Default is "gru".
        bidirectional : bool, optional
            If True, uses bidirectional wrappers for both recurrent and skip recurrent layers. Default is True.
        input_channels : int, optional
            Number of input channels for the 1D convolution used in skip connections. Default is 64.
        skip_steps : int, optional
            Step size and kernel size used in the skip convolution. Determines how many steps are skipped.
            Also determines the multiplier for the number of filters. Default is 4.
        dropout : float, optional
            Dropout rate applied within the recurrent layers. Default is 0.05.
        **kwargs
            Additional keyword arguments passed to the parent class constructor.
        """
        super().__init__(**layer_kwargs(kwargs))

        self.skip_conv = keras.layers.Conv1D(
            filters=input_channels * skip_steps,
            kernel_size=skip_steps,
            strides=skip_steps,
            padding="same",
            name="skip_conv",
        )

        recurrent_constructor = find_recurrent_net(recurrent_type)

        recurrent = recurrent_constructor(
            units=hidden_dim // 2 if bidirectional else hidden_dim,
            dropout=dropout,
        )
        recurrent.name = "recurrent"

        skip_recurrent = recurrent_constructor(
            units=hidden_dim // 2 if bidirectional else hidden_dim,
            dropout=dropout,
        )
        skip_recurrent.name = "skip_recurrent"

        if bidirectional:
            recurrent = keras.layers.Bidirectional(recurrent, name="bidirectional_recurrent")
            skip_recurrent = keras.layers.Bidirectional(skip_recurrent, name="bidirectional_skip_recurrent")

        self.recurrent = recurrent
        self.skip_recurrent = skip_recurrent
        self.input_channels = input_channels

    def call(self, time_series: Tensor, training: bool = False, **kwargs) -> Tensor:
        direct_summary = self.recurrent(time_series, training=training)
        skip_summary = self.skip_recurrent(self.skip_conv(time_series), training=training)
        return keras.ops.concatenate((direct_summary, skip_summary), axis=-1)

    @sanitize_input_shape
    def build(self, input_shape):
        super().build(input_shape)



import os

SWITCH_VERSION = "v1.1.6"

TEMPLATE_LATEST = """<!doctype html>
<html>
    <head>
    <title>Redirecting to {{{{ latest.name }}}} branch/version</title>
        <meta charset="utf-8" />
        <meta
            http-equiv="refresh"
            content="0; url=/{{{{ latest.name }}}}/{path}"
        />
        <link rel="canonical" href="/{{{{ latest.name }}}}/{path}" />
    </head>
</html>
"""

TEMPLATE_V1 = f"""<!doctype html>
<html>
    <head>
    <title>Redirecting to {SWITCH_VERSION}</title>
        <meta charset="utf-8" />
        <meta
            http-equiv="refresh"
            content="0; url=/{SWITCH_VERSION}/{{path}}"
        />
        <link rel="canonical" href="/{SWITCH_VERSION}/{{path}}" />
    </head>
</html>
"""


def generate_templates(old_doc_dir, include_dirs, latest_files, template_dir):
    os.makedirs(template_dir, exist_ok=True)

    for curdir, dirs, files in os.walk(old_doc_dir):
        if any([curdir.startswith(old_doc_dir + "/" + d) for d in include_dirs]) or curdir == old_doc_dir:
            html_files = [f for f in files if f.endswith("html")]
            if html_files:
                os.makedirs(os.path.join(template_dir, curdir), exist_ok=True)
            for file in html_files:
                file_path = os.path.join(curdir[(len(old_doc_dir) + 1) :], file)
                print("Writing", os.path.join(curdir[(len(old_doc_dir) + 1) :], file))
                with open(os.path.join(template_dir, file_path), "w") as f:
                    template = TEMPLATE_LATEST if file in latest_files else TEMPLATE_V1
                    content = template.format(path=file_path)
                    f.write(content)


if __name__ == "__main__":
    old_doc_dir = "../docs/v1.1.6"
    include_dirs = ["_examples", "_images", "_modules", "api"]
    latest_files = ["index.html", "about.html"]
    template_dir = "polyversion/templates"

    generate_templates(old_doc_dir, include_dirs, latest_files, template_dir)



import pytest
import numpy as np

import keras

from bayesflow.utils.serialization import deserialize, serialize

import bayesflow as bf


def test_cycle_consistency(adapter, random_data):
    processed = adapter(random_data)
    deprocessed = adapter(processed, inverse=True)

    for key, value in random_data.items():
        if key in ["d1", "d2"]:
            # dropped
            continue
        assert key in deprocessed
        assert np.allclose(value, deprocessed[key])


def test_serialize_deserialize(adapter, random_data):
    processed = adapter(random_data)
    serialized = serialize(adapter)
    deserialized = deserialize(serialized)
    reserialized = serialize(deserialized)

    assert keras.tree.lists_to_tuples(serialized) == keras.tree.lists_to_tuples(reserialized)

    random_data["foo"] = random_data["x1"]
    deserialized_processed = deserialized(random_data)
    for key, value in processed.items():
        assert np.allclose(value, deserialized_processed[key])


def test_constrain():
    # check if constraint-implied transforms are applied correctly
    import numpy as np
    import warnings
    from bayesflow.adapters import Adapter

    data = {
        "x_lower_cont": np.random.exponential(1, size=(32, 1)),
        "x_upper_cont": -np.random.exponential(1, size=(32, 1)),
        "x_both_cont": np.random.beta(0.5, 0.5, size=(32, 1)),
        "x_lower_disc1": np.zeros(shape=(32, 1)),
        "x_lower_disc2": np.zeros(shape=(32, 1)),
        "x_upper_disc1": np.ones(shape=(32, 1)),
        "x_upper_disc2": np.ones(shape=(32, 1)),
        "x_both_disc1": np.vstack((np.zeros(shape=(16, 1)), np.ones(shape=(16, 1)))),
        "x_both_disc2": np.vstack((np.zeros(shape=(16, 1)), np.ones(shape=(16, 1)))),
    }

    ad = (
        Adapter()
        .constrain("x_lower_cont", lower=0)
        .constrain("x_upper_cont", upper=0)
        .constrain("x_both_cont", lower=0, upper=1)
        .constrain("x_lower_disc1", lower=0, inclusive="lower")
        .constrain("x_lower_disc2", lower=0, inclusive="none")
        .constrain("x_upper_disc1", upper=1, inclusive="upper")
        .constrain("x_upper_disc2", upper=1, inclusive="none")
        .constrain("x_both_disc1", lower=0, upper=1, inclusive="both")
        .constrain("x_both_disc2", lower=0, upper=1, inclusive="none")
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        result = ad(data)

    # continuous variables should not have boundary issues
    assert result["x_lower_cont"].min() < 0.0
    assert result["x_upper_cont"].max() > 0.0
    assert result["x_both_cont"].min() < 0.0
    assert result["x_both_cont"].max() > 1.0

    # discrete variables at the boundaries should not have issues
    # if inclusive is set properly
    assert np.isfinite(result["x_lower_disc1"].min())
    assert np.isfinite(result["x_upper_disc1"].max())
    assert np.isfinite(result["x_both_disc1"].min())
    assert np.isfinite(result["x_both_disc1"].max())

    # discrete variables at the boundaries should have issues
    # if inclusive is not set properly
    assert np.isneginf(result["x_lower_disc2"][0])
    assert np.isinf(result["x_upper_disc2"][0])
    assert np.isneginf(result["x_both_disc2"][0])
    assert np.isinf(result["x_both_disc2"][-1])


def test_simple_transforms(random_data):
    # check if simple transforms are applied correctly
    from bayesflow.adapters import Adapter

    ad = Adapter().log(["p2", "t2"]).log("t1", p1=True).sqrt("p1")

    result = ad(random_data)

    assert np.allclose(result["p2"], np.log(random_data["p2"]))
    assert np.allclose(result["t2"], np.log(random_data["t2"]))
    assert np.allclose(result["t1"], np.log1p(random_data["t1"]))
    assert np.allclose(result["p1"], np.sqrt(random_data["p1"]))

    # inverse results should match the original input
    inverse = ad(result, inverse=True)

    assert np.allclose(inverse["p2"], random_data["p2"])
    assert np.allclose(inverse["t2"], random_data["t2"])
    assert np.allclose(inverse["t1"], random_data["t1"])

    assert np.allclose(inverse["p1"], random_data["p1"])


def test_custom_transform():
    # test that transform raises errors in all relevant cases
    import keras
    from bayesflow.adapters.transforms import SerializableCustomTransform
    from copy import deepcopy

    class A:
        @classmethod
        def fn(cls, x):
            return x

    def not_registered_fn(x):
        return x

    @keras.saving.register_keras_serializable("custom")
    def registered_fn(x):
        return x

    @keras.saving.register_keras_serializable("custom")
    def registered_but_changed(x):
        return x

    def registered_but_changed(x):  # noqa: F811
        return 2 * x

    # method instead of function provided
    with pytest.raises(ValueError):
        SerializableCustomTransform(forward=A.fn, inverse=registered_fn)
    with pytest.raises(ValueError):
        SerializableCustomTransform(forward=registered_fn, inverse=A.fn)

    # lambda function provided
    with pytest.raises(ValueError):
        SerializableCustomTransform(forward=lambda x: x, inverse=registered_fn)
    with pytest.raises(ValueError):
        SerializableCustomTransform(forward=registered_fn, inverse=lambda x: x)

    # unregistered function provided
    with pytest.raises(ValueError):
        SerializableCustomTransform(forward=not_registered_fn, inverse=registered_fn)
    with pytest.raises(ValueError):
        SerializableCustomTransform(forward=registered_fn, inverse=not_registered_fn)

    # function does not match the registered function
    with pytest.raises(ValueError):
        SerializableCustomTransform(forward=registered_but_changed, inverse=registered_fn)
    with pytest.raises(ValueError):
        SerializableCustomTransform(forward=registered_fn, inverse=registered_but_changed)

    transform = SerializableCustomTransform(forward=registered_fn, inverse=registered_fn)
    serialized_transform = keras.saving.serialize_keras_object(transform)
    keras.saving.deserialize_keras_object(serialized_transform)

    # modify name of the forward function so that it cannot be found
    corrupt_serialized_transform = deepcopy(serialized_transform)
    corrupt_serialized_transform["config"]["forward"]["config"] = "nonexistent"
    with pytest.raises(TypeError):
        keras.saving.deserialize_keras_object(corrupt_serialized_transform)

    # modify the name of the inverse transform so that it cannot be found
    corrupt_serialized_transform = deepcopy(serialized_transform)
    corrupt_serialized_transform["config"]["inverse"]["config"] = "nonexistent"
    with pytest.raises(TypeError):
        keras.saving.deserialize_keras_object(corrupt_serialized_transform)


def test_split_transform(adapter, random_data):
    assert "key_to_split" in random_data

    shape = random_data["key_to_split"].shape
    target_shape = (*shape[:-1], shape[-1] // 2)

    processed = adapter(random_data)

    assert "split_1" in processed
    assert processed["split_1"].shape == target_shape

    assert "split_2" in processed
    assert processed["split_2"].shape == target_shape


def test_to_dict_transform():
    import pandas as pd

    data = {
        "int32": [1, 2, 3, 4, 5],
        "int64": [1, 2, 3, 4, 5],
        "float32": [1.0, 2.0, 3.0, 4.0, 5.0],
        "float64": [1.0, 2.0, 3.0, 4.0, 5.0],
        "object": ["a", "b", "c", "d", "e"],
        "category": ["one", "two", "three", "four", "five"],
    }

    df = pd.DataFrame(data)
    df["int32"] = df["int32"].astype("int32")
    df["int64"] = df["int64"].astype("int64")
    df["float32"] = df["float32"].astype("float32")
    df["float64"] = df["float64"].astype("float64")
    df["object"] = df["object"].astype("object")
    df["category"] = df["category"].astype("category")

    ad = bf.Adapter().to_dict()

    # drop one element to simulate non-complete data
    batch = df.iloc[:-1]

    processed = ad(batch)

    assert isinstance(processed, dict)
    assert list(processed.keys()) == ["int32", "int64", "float32", "float64", "object", "category"]

    for key, value in processed.items():
        assert isinstance(value, np.ndarray)
        assert value.dtype == "float32"

    # category should have 5 one-hot categories, even though it was only passed 4
    assert processed["category"].shape[-1] == 5



"""Scoring rules for point estimation."""

from .scoring_rule import ScoringRule
from .parametric_distribution_score import ParametricDistributionScore
from .normed_difference_score import NormedDifferenceScore
from .mean_score import MeanScore
from .median_score import MedianScore
from .quantile_score import QuantileScore
from .multivariate_normal_score import MultivariateNormalScore

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



import numpy as np

from .io import format_bytes, parse_bytes
from . import logging
from .tensor_utils import size_of


def find_batch_size(sample, memory_budget: str = "auto", min: int = 4, max: int = 1024) -> int:
    """Returns an estimation of an optimal batch size based on memory budget and sample memory footprint.

    :param sample: Any nested structure of tensors, representing a single sample.
    :param memory_budget: The maximum available memory for a single batch.
    :param min: The minimum batch size.
    :param max: The maximum batch size.
    """
    if memory_budget == "auto":
        memory_budget = find_memory_budget()
    elif isinstance(memory_budget, str):
        memory_budget = parse_bytes(memory_budget)

    # find the size of one sample
    sample_memory = size_of(sample)

    logging.info(f"Estimating memory footprint of one sample at {format_bytes(sample_memory, precision=1)}.")

    # use a conservative (low) estimate for the optimal batch size
    batch_size = memory_budget / (4 * sample_memory)

    if batch_size < 16:
        logging.warning(
            "Memory budget is very small compared to sample size. You may need to accumulate gradients over "
            f"multiple batches using `gradient_accumulation_steps` in the optimizer. We recommend accumulating "
            f"at least {int(32 / batch_size)} steps."
        )

    # limit estimate to sensible range
    batch_size = int(np.clip(batch_size, min, max))

    return batch_size


def find_memory_budget(device_type: str = None) -> int:
    """Returns an estimation of available memory in bytes for the given device type."""
    # keras utilities for device information are not very mature yet
    raise NotImplementedError("Automatic memory budget is not yet supported. Please pass an explicit value.")



from functools import singledispatch
import keras


@singledispatch
def find_distribution(arg, **kwargs):
    raise TypeError(f"Cannot infer distribution from {arg!r}.")


@find_distribution.register
def _(name: str, *args, **kwargs):
    match name.lower():
        case "normal":
            from bayesflow.distributions import DiagonalNormal

            distribution = DiagonalNormal(*args, **kwargs)

        case "student" | "student-t" | "student_t":
            from bayesflow.distributions import DiagonalStudentT

            distribution = DiagonalStudentT(*args, **kwargs)

        case "mixture":
            raise ValueError(
                "Mixture distributions need to be explicitly defined as bf.distributions.Mixture(...) "
                "and passed to the constructor."
            )
        case "none":
            distribution = None

        case other:
            raise ValueError(f"Unsupported distribution name '{other}'.")

    return distribution


@find_distribution.register
def _(none: None, *args, **kwargs):
    return None


@find_distribution.register
def _(distribution: keras.Layer, *args, **kwargs):
    return distribution



import keras
import numpy as np
import pytest

from bayesflow.utils.serialization import deserialize, serialize

from tests.utils import assert_layers_equal


@pytest.mark.parametrize("automatic", [True, False])
def test_build(automatic, summary_network, random_set):
    if summary_network is None:
        pytest.skip(reason="Nothing to do, because there is no summary network.")

    assert summary_network.built is False

    if automatic:
        summary_network(random_set)
    else:
        summary_network.build(keras.ops.shape(random_set))

    assert summary_network.built is True

    # check the model has variables
    assert summary_network.variables, "Model has no variables."


@pytest.mark.parametrize("automatic", [True, False])
def test_build_functional_api(automatic, summary_network, random_set):
    if summary_network is None:
        pytest.skip(reason="Nothing to do, because there is no summary network.")

    assert summary_network.built is False

    inputs = keras.layers.Input(shape=keras.ops.shape(random_set)[1:])
    outputs = summary_network(inputs)
    model = keras.Model(inputs=inputs, outputs=outputs)

    if automatic:
        model(random_set)
    else:
        model.build(keras.ops.shape(random_set))

    assert model.built is True

    # check the model has variables
    assert summary_network.variables, "Model has no variables."


def test_variable_batch_size(summary_network, random_set):
    if summary_network is None:
        pytest.skip(reason="Nothing to do, because there is no summary network.")

    # build with one batch size
    summary_network.build(keras.ops.shape(random_set))

    # run with another batch size
    batch_sizes = np.random.choice(10, replace=False, size=3)
    for b in batch_sizes:
        new_input = keras.ops.zeros((b,) + keras.ops.shape(random_set)[1:])
        summary_network(new_input)


def test_variable_set_size(summary_network, random_set):
    if summary_network is None:
        pytest.skip(reason="Nothing to do, because there is no summary network.")

    # build with one set size
    summary_network.build(keras.ops.shape(random_set))

    # run with another set size
    for s in [3, 4, 5]:
        b = keras.ops.shape(random_set)[0]
        new_input = keras.ops.zeros((b, s, keras.ops.shape(random_set)[2]))
        summary_network(new_input)


def test_serialize_deserialize(summary_network, random_set):
    if summary_network is None:
        pytest.skip(reason="Nothing to do, because there is no summary network.")

    summary_network.build(keras.ops.shape(random_set))

    serialized = serialize(summary_network)
    deserialized = deserialize(serialized)
    reserialized = serialize(deserialized)

    assert keras.tree.lists_to_tuples(serialized) == keras.tree.lists_to_tuples(reserialized)


def test_save_and_load(tmp_path, summary_network, random_set):
    if summary_network is None:
        pytest.skip(reason="Nothing to do, because there is no summary network.")

    summary_network.build(keras.ops.shape(random_set))

    keras.saving.save_model(summary_network, tmp_path / "model.keras")
    loaded = keras.saving.load_model(tmp_path / "model.keras")

    assert_layers_equal(summary_network, loaded)


@pytest.mark.parametrize("stage", ["training", "validation"])
def test_compute_metrics(stage, summary_network, random_set):
    if summary_network is None:
        pytest.skip("Nothing to do, because there is no summary network.")

    summary_network.build(keras.ops.shape(random_set))

    metrics = summary_network.compute_metrics(random_set, stage=stage)

    assert "outputs" in metrics

    # check that the batch dimension is preserved
    assert keras.ops.shape(metrics["outputs"])[0] == keras.ops.shape(random_set)[0]

    # check summary dimension
    summary_dim = summary_network.summary_dim
    assert keras.ops.shape(metrics["outputs"])[-1] == summary_dim

    if summary_network.base_distribution is not None:
        assert "loss" in metrics
        assert keras.ops.shape(metrics["loss"]) == ()

        if stage != "training":
            for metric in summary_network.metrics:
                assert metric.name in metrics
                assert keras.ops.shape(metrics[metric.name]) == ()



import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform
from .transform import Transform


@serializable
class MapTransform(Transform):
    """
    Implements a transform that applies a set of elementwise transforms
    to a subset of the data as given by a mapping.

    Parameters
    ----------
    transform_map : dict[str, ElementwiseTransform]
        Dictionary with variable names as keys and corresponding transforms as values.
    """

    def __init__(self, transform_map: dict[str, ElementwiseTransform]):
        self.transform_map = transform_map

    def __repr__(self):
        # if the transform map values are all the same type, use that type name
        transform_types = {type(transform) for transform in self.transform_map.values()}

        if len(transform_types) == 1:
            transform_type = transform_types.pop()
        else:
            transform_type = MapTransform

        if e := self.extra_repr():
            return f"{transform_type.__name__}({e})"

        return transform_type.__name__

    def get_config(self) -> dict:
        return serialize({"transform_map": self.transform_map})

    def forward(self, data: dict[str, np.ndarray], *, strict: bool = True, **kwargs) -> dict[str, np.ndarray]:
        data = data.copy()

        required_keys = set(self.transform_map.keys())
        available_keys = set(data.keys())
        missing_keys = required_keys - available_keys

        if strict and missing_keys:
            raise KeyError(f"Missing keys: {missing_keys!r}")

        for key, transform in self.transform_map.items():
            if key in data:
                data[key] = transform.forward(data[key], **kwargs)

        return data

    def inverse(self, data: dict[str, np.ndarray], *, strict: bool = False, **kwargs) -> dict[str, np.ndarray]:
        data = data.copy()

        required_keys = set(self.transform_map.keys())
        available_keys = set(data.keys())
        missing_keys = required_keys - available_keys

        if strict and missing_keys:
            raise KeyError(f"Missing keys: {missing_keys!r}")

        for key, transform in self.transform_map.items():
            if key in data:
                data[key] = transform.inverse(data[key], **kwargs)

        return data



import pytest

from tests.utils import run_notebook


@pytest.mark.skip(reason="requires setting up Stan")
@pytest.mark.slow
def test_bayesian_experimental_design(examples_path):
    run_notebook(examples_path / "Bayesian_Experimental_Design.ipynb")


@pytest.mark.skip(reason="requires setting up pyabc")
@pytest.mark.slow
def test_from_abc_to_bayesflow(examples_path):
    run_notebook(examples_path / "From_ABC_to_BayesFlow.ipynb")


@pytest.mark.slow
def test_linear_regression_starter(examples_path):
    run_notebook(examples_path / "Linear_Regression_Starter.ipynb")


@pytest.mark.slow
def test_lotka_volterra_point_estimation_and_expert_stats(examples_path):
    run_notebook(examples_path / "Lotka_Volterra_Point_Estimation_and_Expert_Stats.ipynb")


@pytest.mark.slow
def test_one_sample_ttest(examples_path):
    run_notebook(examples_path / "One_Sample_TTest.ipynb")


@pytest.mark.slow
def test_sir_posterior_estimation(examples_path):
    run_notebook(examples_path / "SIR_Posterior_Estimation.ipynb")


@pytest.mark.slow
def test_two_moons_starter(examples_path):
    run_notebook(examples_path / "Two_Moons_Starter.ipynb")



import numpy as np

from bayesflow.utils.serialization import serializable, deserialize


@serializable
class Transform:
    """
    Base class on which other transforms are based
    """

    def __call__(self, data: dict[str, np.ndarray], *, inverse: bool = False, **kwargs) -> dict[str, np.ndarray]:
        if inverse:
            return self.inverse(data, **kwargs)

        return self.forward(data, **kwargs)

    def __repr__(self):
        if e := self.extra_repr():
            return f"{self.__class__.__name__}({e})"
        return self.__class__.__name__

    @classmethod
    def from_config(cls, config: dict, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self) -> dict:
        raise NotImplementedError

    def forward(self, data: dict[str, np.ndarray], **kwargs) -> dict[str, np.ndarray]:
        raise NotImplementedError

    def inverse(self, data: dict[str, np.ndarray], **kwargs) -> dict[str, np.ndarray]:
        raise NotImplementedError

    def extra_repr(self) -> str:
        return ""



from collections.abc import Sequence
import numpy as np
from scipy.stats import binom as scipy_binomial


def pointwise_ecdf_bands(
    num_estimates: int,
    num_points: int = None,
    confidence: float = 0.99,
    min_point: float = 1e-5,
    max_point: float = 1 - 1e-5,
    max_num_points: int = 1000,
) -> Sequence:
    """Computes the pointwise ECDF confidence bands from the inverse CDF of the binomial distribution.

    Refer to the following for context and notation:

    Säilynoja, T., Bürkner, P. C., & Vehtari, A. (2022).
    Graphical test for discrete uniformity and its applications in goodness-of-fit
    evaluation and multiple sample comparison. Statistics and Computing, 32(2), 32.
    See: https://link.springer.com/article/10.1007/s11222-022-10090-6

    Will be used by the diagnostics module to create the ECDF marginal calibration plots.

    Parameters
    ----------
    num_estimates   : int
        The sample size used for computing the ECDF. Will equal to the number of simulated conditions when used
        for simulation-based calibration.
    num_points      : int, optional, default: None
        The number of evaluation points on the interval (0, 1). Defaults to `num_points = num_estimates` if
        not explicitly specified. Correspond to `K` in the paper above.
    confidence      : float in (0, 1), optional, default: 0.95
        The confidence level, `confidence = 1 - alpha` specifies the width of the confidence interval.
    min_point       : float, optional, default: 1e-5
        Lower bound of the interval for which the confidence bands are computed.
    max_point       : float, optional, default: 1-1e-5
        Upper bound of the interval for which the confidence bands are computed.
    max_num_points  : int, optional, default: 1000
        Upper bound on `num_points`. Saves computation time when `num_estimates` is large.

    Returns
    -------
    (alpha, z, L, U) - tuple of scalar and three arrays of size (num_estimates,) containing the confidence level
        as well as the evaluation points, the lower, and the upper confidence bands, respectively.
    """

    N = num_estimates
    if num_points is None:
        K = min(N, max_num_points)
    else:
        K = min(num_points, max_num_points)
    z = np.linspace(min_point, max_point, K)

    alpha = 1 - confidence
    L = np.zeros_like(z)
    U = np.zeros_like(z)

    for i, p in enumerate(z):
        L[i] = scipy_binomial.ppf(alpha / 2, N, p) / N
        U[i] = scipy_binomial.ppf(1 - alpha / 2, N, p) / N

    return alpha, z, L, U



from functools import singledispatch
import keras


@singledispatch
def find_summary_network(arg, *args, **kwargs):
    raise TypeError(f"Cannot infer inference network from {arg!r}.")


@find_summary_network.register
def _(name: str, *args, **kwargs):
    match name.lower():
        case "deep_set":
            from bayesflow.networks import DeepSet

            return DeepSet(*args, **kwargs)

        case "set_transformer":
            from bayesflow.networks import SetTransformer

            return SetTransformer(*args, **kwargs)

        case "fusion_transformer":
            from bayesflow.networks import FusionTransformer

            return FusionTransformer(*args, **kwargs)

        case "time_series_transformer":
            from bayesflow.networks import TimeSeriesTransformer

            return TimeSeriesTransformer(*args, **kwargs)

        case "time_series_network":
            from bayesflow.networks import TimeSeriesNetwork

            return TimeSeriesNetwork(*args, **kwargs)

        case unknown_network:
            raise ValueError(f"Unknown summary network: '{unknown_network}'")


@find_summary_network.register
def _(layer: keras.Layer, *args, **kwargs):
    return layer


@find_summary_network.register
def _(model: keras.Model, *args, **kwargs):
    return model



from collections.abc import Sequence

import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform


@serializable
class Standardize(ElementwiseTransform):
    """
    Transform that when applied standardizes data using typical z-score standardization
    i.e. for some unstandardized data x the standardized version z would be

    >>> z = (x - mean(x)) / std(x)

    Parameters
    ----------
    mean : int or float, optional
        Specify a mean if known but will be estimated from data when not provided
    std : int or float, optional
        Specify a standard devation if known but will be estimated from data when not provided
    axis : int, optional
        A specific axis along which standardization should take place. By default
        standardization happens individually for each dimension
    momentum : float in (0,1)
        The momentum during training

    Examples
    --------
    1) Standardize all variables using their individually estimated mean and stds.

    >>> adapter = (
            bf.adapters.Adapter()
                .standardize()
        )


    2) Standardize all with same known mean and std.

    >>> adapter = (
            bf.adapters.Adapter()
                .standardize(mean = 5, sd = 10)
        )


    3) Mix of fixed and estimated means/stds. Suppose we have priors for "beta" and "sigma" where we
    know the means and stds. However for all other variables, the means and stds are unknown.
    Then standardize should be used in several stages specifying which variables to include or exclude.

    >>> adapter = (
            bf.adapters.Adapter()
                # mean fixed, std estimated
                .standardize(include = "beta", mean = 1)
                # both mean and SD fixed
                .standardize(include = "sigma", mean = 0.6, sd = 3)
                # both means and stds estimated for all other variables
                .standardize(exclude = ["beta", "sigma"])
        )
    """

    def __init__(
        self,
        mean: int | float | np.ndarray = None,
        std: int | float | np.ndarray = None,
        axis: int | Sequence[int] = None,
        momentum: float | None = 0.99,
    ):
        super().__init__()

        self.mean = mean
        self.std = std

        if isinstance(axis, Sequence):
            # numpy hates lists
            axis = tuple(axis)
        self.axis = axis
        self.momentum = momentum

    def get_config(self) -> dict:
        config = {
            "mean": self.mean,
            "std": self.std,
            "axis": self.axis,
            "momentum": self.momentum,
        }
        return serialize(config)

    def forward(self, data: np.ndarray, stage: str = "inference", **kwargs) -> np.ndarray:
        if self.axis is None:
            self.axis = tuple(range(data.ndim - 1))

        if self.mean is None:
            self.mean = np.mean(data, axis=self.axis, keepdims=True)
        else:
            if self.momentum is not None and stage == "training":
                self.mean = self.momentum * self.mean + (1.0 - self.momentum) * np.mean(
                    data, axis=self.axis, keepdims=True
                )

        if self.std is None:
            self.std = np.std(data, axis=self.axis, keepdims=True, ddof=1)
        else:
            if self.momentum is not None and stage == "training":
                self.std = self.momentum * self.std + (1.0 - self.momentum) * np.std(
                    data, axis=self.axis, keepdims=True, ddof=1
                )

        mean = np.broadcast_to(self.mean, data.shape)
        std = np.broadcast_to(self.std, data.shape)

        return (data - mean) / std

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        if self.mean is None or self.std is None:
            raise RuntimeError("Cannot call `inverse` before calling `forward` at least once.")

        mean = np.broadcast_to(self.mean, data.shape)
        std = np.broadcast_to(self.std, data.shape)

        return data * std + mean



from collections.abc import Callable, Sequence

import numpy as np

from bayesflow.utils import batched_call, filter_kwargs, tree_stack
from bayesflow.utils.decorators import allow_batch_size

from .simulator import Simulator
from ..types import Shape


class LambdaSimulator(Simulator):
    """Implements a simulator based on a sampling function."""

    def __init__(self, sample_fn: Callable[[Sequence[int]], dict[str, any]], *, is_batched: bool = False):
        """
        Initialize a simulator based on a simple callable function

        Parameters
        ----------
        sample_fn : Callable[[Sequence[int]], dict[str, any]]
            A function that generates samples. It should accept `batch_shape` as its first argument
            (if `is_batched=True`), followed by keyword arguments.
        is_batched : bool, optional
            Whether the `sample_fn` is implemented to handle batched sampling directly.
            If False, `sample_fn` will be called once per sample and results will be stacked.
            Default is False.
        """
        self.sample_fn = sample_fn
        self.is_batched = is_batched

    @allow_batch_size
    def sample(self, batch_shape: Shape, **kwargs) -> dict[str, np.ndarray]:
        """
        Sample using the wrapped sampling function.

        Parameters
        ----------
        batch_shape : Shape
            The shape of the batch to sample. Typically, a tuple indicating the number of samples,
            but an int can also be passed.
        **kwargs
            Additional keyword arguments passed to the sampling function. Only valid arguments
            (as determined by the function's signature) are used.

        Returns
        -------
        data : dict of str to np.ndarray
            A dictionary of sampled outputs. Keys are output names and values are numpy arrays.
            If `is_batched` is False, individual outputs are stacked along the first axis.
        """

        # try to use only valid keyword-arguments
        kwargs = filter_kwargs(kwargs, self.sample_fn)

        if self.is_batched:
            return self.sample_fn(batch_shape, **kwargs)

        data = batched_call(self.sample_fn, batch_shape, kwargs=kwargs, flatten=True)
        data = tree_stack(data, axis=0, numpy=True)

        return data



from bayesflow.utils.serialization import serializable, serialize

from .transform import Transform


@serializable
class Rename(Transform):
    """
    Transform to rename keys in data dictionary. Useful to rename variables to match those required by
    approximator. This transform can only rename one variable at a time.

    Parameters
    ----------
    from_key : str
        Variable name that should be renamed
    to_key : str
        New variable name

    Examples
    --------
    >>> adapter = (
        bf.adapters.Adapter()
        # rename the variables to match the required approximator inputs
        .rename("theta", "inference_variables")
        .rename("x", "inference_conditions")
    )
    """

    def __init__(self, from_key: str, to_key: str):
        super().__init__()
        self.from_key = from_key
        self.to_key = to_key

    def get_config(self) -> dict:
        return serialize({"from_key": self.from_key, "to_key": self.to_key})

    def forward(self, data: dict[str, any], *, strict: bool = True, **kwargs) -> dict[str, any]:
        data = data.copy()

        if strict and self.from_key not in data:
            raise KeyError(f"Missing key: {self.from_key!r}")
        elif self.from_key not in data:
            return data

        data[self.to_key] = data.pop(self.from_key)
        return data

    def inverse(self, data: dict[str, any], *, strict: bool = False, **kwargs) -> dict[str, any]:
        data = data.copy()

        if strict and self.to_key not in data:
            raise KeyError(f"Missing key: {self.to_key!r}")
        elif self.to_key not in data:
            return data

        data[self.from_key] = data.pop(self.to_key)
        return data

    def extra_repr(self) -> str:
        return f"{self.from_key!r} -> {self.to_key!r}"



from datetime import datetime
import logging
from pathlib import Path
from docsrc.polyversion_patches import DynamicPip, CustomDriver, PyDataVersionEncoder

from sphinx_polyversion.api import apply_overrides
from sphinx_polyversion.git import Git, GitRef, GitRefType, file_predicate, refs_by_type
from sphinx_polyversion.pyvenv import Environment, VenvWrapper
from sphinx_polyversion.sphinx import SphinxBuilder, Placeholder


logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

#: Determine repository root directory
root = Git.root(Path(__file__).parent)

#: CodeRegex matching the branches to build docs for
BRANCH_REGEX = r"^(main|stable-legacy)$"

#: Regex matching the tags to build docs for
# TAG_REGEX = r"^v((?!1\.)[\.0-9]*)$"
TAG_REGEX = r""

#: Output dir relative to project root
OUTPUT_DIR = "_build_polyversion"

#: Source directory
SOURCE_DIR = "docsrc"

MOCK_DATA = {
    "revisions": [],
    "current": GitRef("local", "", "", GitRefType.BRANCH, datetime.now()),
}

#: Whether to build using only local files and mock data
MOCK = False

#: Whether to run the builds in sequence or in parallel
SEQUENTIAL = False

#: Arguments to pass to `sphinx-build`
SPHINX_ARGS = "-a -v"

#: Extra packages for building docs
SPHINX_DEPS = [
    "sphinx",
    "numpydoc",
    "myst-nb",
    "sphinx_design",
    "pydata-sphinx-theme",
    "sphinxcontrib-bibtex",
    "sphinx-polyversion==1.1.0",
]

#: Extra dependencies to iinstall for version 1
V1_BACKEND_DEPS = [
    "tf-keras",
]

#: Extra dependencies to install for version 2
V2_BACKEND_DEPS = [
    "jax",
    # "torch",
    # "tensorflow",
]

VENV_DIR_NAME = ".docs_venvs"


#: Data passed to templates
def data(driver, rev, env):
    revisions = driver.targets
    branches, tags = refs_by_type(revisions)
    latest = max(tags or branches)
    for b in branches:
        if b.name == "main":
            latest = b
            break

    # sort tags and branches by date, newest first
    return {
        "current": rev,
        "tags": sorted(tags, reverse=True),
        "branches": sorted(branches, reverse=True),
        "revisions": revisions,
        "latest": latest,
    }


def root_data(driver):
    revisions = driver.builds
    branches, tags = refs_by_type(revisions)
    latest = max(tags or branches)
    for b in branches:
        if b.name == "main":
            latest = b
            break
    return {"revisions": revisions, "latest": latest}


# Load overrides read from commandline to global scope
apply_overrides(globals())


# Setup environments for the different versions
src = Path(SOURCE_DIR)
vcs = Git(
    branch_regex=BRANCH_REGEX,
    tag_regex=TAG_REGEX,
    buffer_size=1 * 10**9,  # 1 GB
    predicate=file_predicate([src]),  # exclude refs without source dir
)


creator = VenvWrapper(with_pip=True)


async def selector(rev, keys):
    """Select configuration based on revision"""
    # map all v1 revisions to one configuration
    if rev.name.startswith("v1.") or rev.name == "stable-legacy":
        return "stable-legacy"
    elif rev.name == "local":
        return "local"
    # common config for everything else
    return None


shared_env_kwargs = dict(temporary=SEQUENTIAL, creator=creator, venv=Path(VENV_DIR_NAME))
ENVIRONMENT = {
    # configuration for v2 and dev
    None: DynamicPip.factory(
        **shared_env_kwargs,
        args=["-e", "."] + SPHINX_DEPS + V2_BACKEND_DEPS,
        env={"KERAS_BACKEND": "jax"},
    ),
    # configuration for v1
    "stable-legacy": DynamicPip.factory(
        **shared_env_kwargs,
        args=["-e", "."] + SPHINX_DEPS + V1_BACKEND_DEPS,
        env={"SETUPTOOLS_SCM_PRETEND_VERSION_FOR_BAYESFLOW": "1.1.6"},
    ),
    # use local envorinment for local build
    "local": Environment.factory(),
}


# Setup driver and run it
CustomDriver(
    root,
    OUTPUT_DIR,
    vcs=vcs,
    builder=SphinxBuilder(
        src / "source",
        args=SPHINX_ARGS.split(),
        pre_cmd=["python", root / src / "pre-build.py", Placeholder.SOURCE_DIR],
    ),
    env=ENVIRONMENT,
    encoder=PyDataVersionEncoder(),
    selector=selector,
    template_dir=root / src / "polyversion/templates",
    static_dir=root / src / "polyversion/static",
    data_factory=data,
    root_data_factory=root_data,
    mock=MOCK_DATA,
).run(MOCK, SEQUENTIAL)



import keras
from typing import TypeVar

match keras.backend.backend():
    case "numpy":
        import numpy as np

        BackendTensor = np.ndarray
    case "jax":
        import jax

        BackendTensor = jax.Array
    case "tensorflow":
        import tensorflow as tf

        BackendTensor = tf.Tensor
    case "torch":
        import torch

        BackendTensor = torch.Tensor
    case other:
        raise NotImplementedError

Tensor = TypeVar("Tensor", bound=BackendTensor)



import nbformat
from nbconvert.preprocessors import ExecutePreprocessor

from pathlib import Path
import shutil


def run_notebook(path):
    path = Path(path)
    checkpoint_path = path.parent / "checkpoints"
    # only clean up if the directory did not exist before the test
    cleanup_checkpoints = not checkpoint_path.exists()
    with open(str(path), encoding="utf-8") as f:
        nb = nbformat.read(f, nbformat.NO_CONVERT)

    kernel = ExecutePreprocessor(timeout=3600, kernel_name="python3", resources={"metadata": {"path": path.parent}})

    try:
        result = kernel.preprocess(nb)
    finally:
        if cleanup_checkpoints and checkpoint_path.exists():
            # clean up if the directory was created by the test
            shutil.rmtree(checkpoint_path)

    return result



import math

import numpy as np

import keras
from keras import ops

from bayesflow.types import Shape, Tensor
from bayesflow.utils import expand_tile
from bayesflow.utils.decorators import allow_batch_size
from bayesflow.utils.serialization import serializable, serialize

from .distribution import Distribution


@serializable
class DiagonalStudentT(Distribution):
    """Implements a backend-agnostic diagonal Student-t distribution."""

    def __init__(
        self,
        df: int | float,
        loc: int | float | np.ndarray | Tensor = 0.0,
        scale: int | float | np.ndarray | Tensor = 1.0,
        trainable_parameters: bool = False,
        seed_generator: keras.random.SeedGenerator = None,
        **kwargs,
    ):
        """
        Initializes a backend-agnostic Student's t-distribution with optional learnable parameters.

        This class represents a Student's t-distribution, which is useful for modeling heavy-tailed data.
        The distribution is parameterized by degrees of freedom (`df`), location (`loc`), and scale (`scale`).
        These parameters can either be fixed or learned during training.

        The class also supports random number generation with an optional seed for reproducibility.

        Parameters
        ----------
        df : int or float
            Degrees of freedom for the Student's t-distribution. Lower values result in
            heavier tails, making it more robust to outliers.
        loc : int, float, np.ndarray, or Tensor, optional
            The location parameter (mean) of the distribution. Default is 0.0.
        scale : int, float, np.ndarray, or Tensor, optional
            The scale parameter (standard deviation) of the distribution. Default is 1.0.
        trainable_parameters : bool, optional
            Whether to treat `loc` and `scale` as trainable parameters. Default is False.
        seed_generator : keras.random.SeedGenerator, optional
            A Keras seed generator for reproducible random sampling. If None, a new seed
            generator is created. Default is None.
        **kwargs
            Additional keyword arguments passed to the base `Distribution` class.
        """

        super().__init__(**kwargs)

        self.df = df
        self.loc = loc
        self.scale = scale

        self.trainable_parameters = trainable_parameters

        self.seed_generator = seed_generator or keras.random.SeedGenerator()

        self.log_normalization_constant = None
        self.dim = None
        self._loc = None
        self._scale = None

    def build(self, input_shape: Shape) -> None:
        if self.built:
            return

        self.dim = int(input_shape[-1])

        # convert to tensor and broadcast if necessary
        self.loc = ops.cast(ops.broadcast_to(self.loc, (self.dim,)), "float32")
        self.scale = ops.cast(ops.broadcast_to(self.scale, (self.dim,)), "float32")

        self.log_normalization_constant = (
            -0.5 * self.dim * math.log(self.df)
            - 0.5 * self.dim * math.log(math.pi)
            - math.lgamma(0.5 * self.df)
            + math.lgamma(0.5 * (self.df + self.dim))
            - ops.sum(keras.ops.log(self.scale))
        )

        if self.trainable_parameters:
            self._loc = self.add_weight(
                shape=ops.shape(self.loc), initializer=keras.initializers.get(self.loc), dtype="float32", trainable=True
            )
            self._scale = self.add_weight(
                shape=ops.shape(self.scale),
                initializer=keras.initializers.get(self.scale),
                dtype="float32",
                trainable=True,
            )
        else:
            self._loc = self.loc
            self._scale = self.scale

    def log_prob(self, samples: Tensor, *, normalize: bool = True) -> Tensor:
        mahalanobis_term = ops.sum((samples - self._loc) ** 2 / self._scale**2, axis=-1)
        result = -0.5 * (self.df + self.dim) * ops.log1p(mahalanobis_term / self.df)

        if normalize:
            result += self.log_normalization_constant

        return result

    @allow_batch_size
    def sample(self, batch_shape: Shape) -> Tensor:
        # As of writing this code, keras does not support the chi-square distribution
        # nor does it support a scale or rate parameter in Gamma. Hence, we use the relation:
        # chi-square(df) = Gamma(shape = 0.5 * df, scale = 2) = Gamma(shape = 0.5 * df, scale = 1) * 2
        chi2_samples = keras.random.gamma(batch_shape, alpha=0.5 * self.df, seed=self.seed_generator) * 2.0

        # The chi-quare samples need to be repeated across self.dim
        # since for each element of batch_shape only one sample is created.
        chi2_samples = expand_tile(chi2_samples, n=self.dim, axis=-1)

        normal_samples = keras.random.normal(batch_shape + (self.dim,), seed=self.seed_generator)

        return self._loc + self._scale * normal_samples * ops.sqrt(self.df / chi2_samples)

    def get_config(self):
        base_config = super().get_config()

        config = {
            "df": self.df,
            "loc": self.loc,
            "scale": self.scale,
            "trainable_parameters": self.trainable_parameters,
            "seed_generator": self.seed_generator,
        }

        return base_config | serialize(config)



import keras


def devices() -> list:
    """Returns a list of available GPU devices."""
    match keras.backend.backend():
        case "jax":
            import jax

            return jax.devices("gpu")
        case "tensorflow":
            import tensorflow as tf

            return tf.config.list_physical_devices("GPU")
        case "torch":
            import torch

            return [torch.device(f"cuda:{i}") for i in range(torch.cuda.device_count())]
        case "numpy":
            return []
        case _:
            raise NotImplementedError(f"Backend {keras.backend.backend()} not supported.")



import keras

from bayesflow.types import Tensor

from .. import logging

from .euclidean import euclidean


def sinkhorn(x1: Tensor, x2: Tensor, seed: int = None, **kwargs) -> (Tensor, Tensor):
    """
    Matches elements from x2 onto x1 using the Sinkhorn-Knopp algorithm.

    Sinkhorn-Knopp is an iterative algorithm that repeatedly normalizes the cost matrix into a doubly stochastic
    transport plan, containing assignment probabilities.
    The permutation is then sampled randomly according to the transport plan.

    :param x1: Tensor of shape (n, ...)
        Samples from the first distribution.

    :param x2: Tensor of shape (m, ...)
        Samples from the second distribution.

    :param kwargs:
        Additional keyword arguments that are passed to :py:func:`sinkhorn_plan`.

    :param seed: Random seed to use for sampling indices.
        Default: None, which means the seed will be auto-determined for non-compiled contexts.

    :return: Tensor of shape (m,)
        Assignment indices for x2.

    """
    plan = sinkhorn_plan(x1, x2, **kwargs)
    assignments = keras.random.categorical(plan, num_samples=1, seed=seed)
    assignments = keras.ops.squeeze(assignments, axis=1)

    return assignments


def sinkhorn_plan(
    x1: Tensor,
    x2: Tensor,
    regularization: float = 1.0,
    max_steps: int = 10_000,
    rtol: float = 1e-5,
    atol: float = 1e-8,
) -> Tensor:
    """
    Computes the Sinkhorn-Knopp optimal transport plan.

    :param x1: Tensor of shape (n, ...)
        Samples from the first distribution.

    :param x2: Tensor of shape (m, ...)
        Samples from the second distribution.

    :param regularization: Regularization parameter.
        Controls the standard deviation of the Gaussian kernel.

    :param max_steps: Maximum number of iterations, or None to run until convergence.
        Default: 10_000

    :param rtol: Relative tolerance for convergence.
        Default: 1e-5.

    :param atol: Absolute tolerance for convergence.
        Default: 1e-8.

    :return: Tensor of shape (n, m)
        The transport probabilities.
    """
    cost = euclidean(x1, x2)

    # initialize the transport plan from a gaussian kernel
    plan = keras.ops.exp(cost / -(regularization * keras.ops.mean(cost) + 1e-16))

    def contains_nans(plan):
        return keras.ops.any(keras.ops.isnan(plan))

    def is_converged(plan):
        # for convergence, the plan should be doubly stochastic
        conv0 = keras.ops.all(keras.ops.isclose(keras.ops.sum(plan, axis=0), 1.0, rtol=rtol, atol=atol))
        conv1 = keras.ops.all(keras.ops.isclose(keras.ops.sum(plan, axis=1), 1.0, rtol=rtol, atol=atol))
        return conv0 & conv1

    def cond(_, plan):
        # break the while loop if the plan contains nans or is converged
        return ~(contains_nans(plan) | is_converged(plan))

    def body(steps, plan):
        # Sinkhorn-Knopp: repeatedly normalize the transport plan along each dimension
        plan = keras.ops.softmax(plan, axis=0)
        plan = keras.ops.softmax(plan, axis=1)

        return steps + 1, plan

    steps = 0
    steps, plan = keras.ops.while_loop(cond, body, (steps, plan), maximum_iterations=max_steps)

    def do_nothing():
        pass

    def log_steps():
        msg = "Sinkhorn-Knopp converged after {} steps."

        logging.debug(msg, max_steps)

    def warn_convergence():
        msg = "Sinkhorn-Knopp did not converge after {}."

        logging.warning(msg, max_steps)

    def warn_nans():
        msg = "Sinkhorn-Knopp produced NaNs after {} steps."
        logging.warning(msg, steps)

    keras.ops.cond(contains_nans(plan), warn_nans, do_nothing)
    keras.ops.cond(is_converged(plan), log_steps, warn_convergence)

    return plan



from collections.abc import Callable
import keras

from bayesflow.types import Tensor


def jvp(
    f: Callable, x: Tensor | tuple[Tensor, ...], tangents: Tensor | tuple[Tensor, ...], return_output: bool = False
):
    """Compute the Jacobian-vector product of f at x with tangents."""
    if keras.ops.is_tensor(x):
        x = (x,)

    if keras.ops.is_tensor(tangents):
        tangents = (tangents,)

    match keras.backend.backend():
        case "torch":
            import torch

            fx, _jvp = torch.autograd.functional.jvp(f, x, tangents)
        case "tensorflow":
            import tensorflow as tf

            with tf.autodiff.ForwardAccumulator(primals=x, tangents=tangents) as acc:
                fx = f(*x)

            _jvp = acc.jvp(fx)
        case "jax":
            import jax

            fx, _jvp = jax.jvp(
                f,
                x,
                tangents,
            )
        case _:
            raise NotImplementedError(f"JVP not implemented for backend {keras.backend.backend()!r}")

    if return_output:
        return fx, _jvp

    return _jvp



from keras import ops

from bayesflow.types import Shape, Tensor
from bayesflow.utils.serialization import serializable

from .invertible_layer import InvertibleLayer


@serializable
class ActNorm(InvertibleLayer):
    """Implements an Activation Normalization (ActNorm) Layer. Activation Normalization is learned invertible
    normalization, using a scale (s) and a bias (b) vector::

       y = s * x + b(forward)
       x = (y - b) / s(inverse)

    References
    ----------

    [1] Kingma, D. P., & Dhariwal, P. (2018). Glow: Generative flow with invertible 1x1 convolutions.
        Advances in Neural Information Processing Systems, 31.

    [2] Salimans, Tim, and Durk P. Kingma. (2016). Weight normalization: A simple reparameterization to accelerate
       training of deep neural networks. Advances in Neural Information Processing Systems, 29.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.scale = None
        self.bias = None

    def build(self, xz_shape: Shape, **kwargs):
        self.scale = self.add_weight(shape=(xz_shape[-1],), initializer="ones", name="scale")
        self.bias = self.add_weight(shape=(xz_shape[-1],), initializer="zeros", name="bias")

    def call(self, xz: Tensor, inverse: bool = False, **kwargs) -> (Tensor, Tensor):
        if inverse:
            return self._inverse(xz, **kwargs)
        return self._forward(xz, **kwargs)

    def _forward(self, x: Tensor, **kwargs) -> (Tensor, Tensor):
        z = self.scale * x + self.bias
        log_det = ops.sum(ops.log(ops.abs(self.scale)), axis=-1)
        log_det = ops.broadcast_to(log_det, ops.shape(x)[:-1])
        return z, log_det

    def _inverse(self, z: Tensor, **kwargs) -> (Tensor, Tensor):
        x = (z - self.bias) / self.scale
        log_det = -ops.sum(ops.log(ops.abs(self.scale)), axis=-1)
        log_det = ops.broadcast_to(log_det, ops.shape(z)[:-1])
        return x, log_det



import keras
from keras.saving import register_keras_serializable as serializable

from bayesflow.types import Shape, Tensor
from bayesflow.utils import weighted_mean

from .scoring_rule import ScoringRule


@serializable(package="bayesflow.scores")
class NormedDifferenceScore(ScoringRule):
    r""":math:`S(\hat \theta, \theta; k) = | \hat \theta - \theta |^k`

    Scores a point estimate with the k-norm of the error.
    """

    def __init__(self, k: int, **kwargs):
        super().__init__(**kwargs)

        #: Exponent to absolute difference
        self.k = k

        self.config = {"k": k}

    def get_head_shapes_from_target_shape(self, target_shape: Shape):
        # keras.saving.load_model sometimes passes target_shape as a list, so we force a conversion
        target_shape = tuple(target_shape)
        return dict(value=target_shape[1:])

    def score(self, estimates: dict[str, Tensor], targets: Tensor, weights: Tensor = None) -> Tensor:
        r"""
        Computes the scoring function based on the absolute difference between **estimates** and **targets**.

        :math:`S(\hat \theta, \theta; k) = | \hat \theta - \theta |^k`

        This function extracts the Tensor named ``"value"`` from the **estimates** dictionary and computes
        the element-wise absolute difference between the estimates and the true targets. The
        difference is then exponentiated by :py:attr:`k`. The final score is computed using the
        :py:func:`aggregate()` method, which optionally applies weighting.

        Parameters
        ----------
        estimates : dict[str, Tensor]
            A dictionary containing tensors of estimated values. The "value" key must be present.
        targets : Tensor
            A tensor of true target values.
        weights : Tensor, optional
            A tensor of weights corresponding to each estimate-target pair. If provided, it is used
            to compute a weighted aggregate score.

        Returns
        -------
        Tensor
            The aggregated score based on the element-wise absolute difference raised to the power
            of `self.k`, optionally weighted.
        """
        estimates = estimates["value"]
        scores = keras.ops.absolute(estimates - targets) ** self.k
        score = weighted_mean(scores, weights)
        return score

    def get_config(self):
        base_config = super().get_config()
        return base_config | self.config



import numpy as np


def fractional_ranks(post_samples: np.ndarray, prior_samples: np.ndarray) -> np.ndarray:
    """Compute fractional ranks (using broadcasting)"""
    return np.mean(post_samples < prior_samples[:, np.newaxis, :], axis=1)


def _helper_distance_ranks(
    post_samples: np.ndarray,
    prior_samples: np.ndarray,
    stacked: bool,
    references: np.ndarray,
    distance: callable,
    p_norm: int,
) -> np.ndarray:
    """
    Helper function to compute ranks of true parameter wrt posterior samples
    based on distances (defined on the p_norm) between samples and a given references.
    """
    if distance is None:
        # compute distances to references
        dist_post = np.abs((references[:, np.newaxis, :] - post_samples))
        dist_prior = np.abs(references - prior_samples)

        if stacked:
            # compute ranks for all parameters jointly
            samples_distances = np.sum(dist_post**p_norm, axis=-1) ** (1 / p_norm)
            theta_distances = np.sum(dist_prior**p_norm, axis=-1) ** (1 / p_norm)

            ranks = np.mean((samples_distances < theta_distances[:, np.newaxis]), axis=1)[:, np.newaxis]
        else:
            # compute marginal ranks for each parameter
            ranks = np.mean((dist_post < dist_prior[:, np.newaxis]), axis=1)

    else:
        # compute distances using the given distance function
        if stacked:
            # compute distance over joint parameters
            dist_post = np.array([distance(post_samples[i], references[i]) for i in range(references.shape[0])])
            dist_prior = np.array([distance(prior_samples[i], references[i]) for i in range(references.shape[0])])
            ranks = np.mean((dist_post < dist_prior[:, np.newaxis]), axis=1)[:, np.newaxis]
        else:
            # compute distances per parameter
            dist_post = np.zeros_like(post_samples)
            dist_prior = np.zeros_like(prior_samples)
            for i in range(references.shape[0]):  # Iterate over samples
                for j in range(references.shape[1]):  # Iterate over parameters
                    dist_post[i, :, j] = distance(post_samples[i, :, j], references[i, j])
                    dist_prior[i, j] = distance(prior_samples[i, j], references[i, j])

            ranks = np.mean((dist_post < dist_prior[:, np.newaxis]), axis=1)
    return ranks


def distance_ranks(
    post_samples: np.ndarray,
    prior_samples: np.ndarray,
    stacked: bool,
    references: np.ndarray = None,
    distance: callable = None,
    p_norm: int = 2,
) -> np.ndarray:
    """
    Compute ranks of true parameter wrt posterior samples based on distances between samples and optional references.

    Parameters
    ----------
    post_samples : np.ndarray
        The posterior samples.
    prior_samples : np.ndarray
        The prior samples.
    references : np.ndarray, optional
        The references to compute the ranks.
    stacked : bool
        If True, compute ranks for all parameters jointly. Otherwise, compute marginal ranks.
    distance : callable, optional
        The distance function to compute the ranks. If None, the distance defined by the p_norm is used. Must be
        a function that takes two arrays (if stacked, it gets the full parameter vectors, if not only the single
        parameters) and returns an array with the distances. This could be based on the log-posterior, for example.
    p_norm : int, optional
        The norm to compute the distance if no distance is passed. Default is L2-norm.
    """
    # Reference is the origin
    if references is None:
        references = np.zeros((prior_samples.shape[0], prior_samples.shape[1]))
    else:
        # Validate reference
        if references.shape[0] != prior_samples.shape[0]:
            raise ValueError("The number of references must match the number of prior samples.")
        if references.shape[1] != prior_samples.shape[1]:
            raise ValueError("The dimension of references must match the dimension of the parameters.")

    ranks = _helper_distance_ranks(
        post_samples=post_samples,
        prior_samples=prior_samples,
        stacked=stacked,
        references=references,
        distance=distance,
        p_norm=p_norm,
    )
    return ranks



from collections.abc import Sequence, Mapping

import matplotlib.colors
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import numpy as np

from ...utils.plot_utils import make_figure
from ...utils.classification import confusion_matrix


def mc_confusion_matrix(
    pred_models: Mapping[str, np.ndarray] | np.ndarray,
    true_models: Mapping[str, np.ndarray] | np.ndarray,
    model_names: Sequence[str] = None,
    fig_size: tuple = (5, 5),
    label_fontsize: int = 16,
    title_fontsize: int = 18,
    value_fontsize: int = 10,
    tick_fontsize: int = 12,
    xtick_rotation: int = None,
    ytick_rotation: int = None,
    normalize: str = None,
    cmap: matplotlib.colors.Colormap | str = None,
    title: bool = True,
) -> plt.Figure:
    """Plots a confusion matrix for validating a neural network trained for Bayesian model comparison.

    Parameters
    ----------
    pred_models    : np.ndarray of shape (num_data_sets, num_models)
        The predicted posterior model probabilities (PMPs) per data set.
    true_models    : np.ndarray of shape (num_data_sets, num_models)
        The one-hot-encoded true model indices per data set.
    model_names    : list or None, optional, default: None
        The model names for nice plot titles. Inferred if None.
    fig_size       : tuple or None, optional, default: (5, 5)
        The figure size passed to the ``matplotlib`` constructor. Inferred if ``None``
    label_fontsize    : int, optional, default: 16
        The font size of the y-label and y-label texts
    title_fontsize : int, optional, default: 18
        The font size of the title text.
    value_fontsize  : int, optional, default: 10
        The font size of the text annotations and the colorbar tick labels.
    tick_fontsize  : int, optional, default: 12
        The font size of the axis label and model name texts.
    xtick_rotation: int, optional, default: None
        Rotation of x-axis tick labels (helps with long model names).
    ytick_rotation: int, optional, default: None
        Rotation of y-axis tick labels (helps with long model names).
    normalize : {'true', 'pred', 'all'}, default=None
        Passed to confusion matrix. Normalizes confusion matrix over the true (rows),
        predicted (columns) conditions or all the population. If None, confusion matrix
        will not be normalized.
    cmap           : matplotlib.colors.Colormap or str, optional, default: None
        Colormap to be used for the cells. If a str, it should be the name of a registered colormap,
        e.g., 'viridis'. Default colormap matches the BayesFlow defaults by ranging from white to red.
    title          : bool, optional, default True
        A flag for adding 'Confusion Matrix' above the matrix.

    Returns
    -------
    fig : plt.Figure - the figure instance for optional saving
    """

    if model_names is None:
        num_models = true_models.shape[-1]
        model_names = [rf"$M_{{{m}}}$" for m in range(1, num_models + 1)]

    if cmap is None:
        cmap = LinearSegmentedColormap.from_list("", ["white", "#132a70"])

    # Flatten input
    true_models = np.argmax(true_models, axis=1)
    pred_models = np.argmax(pred_models, axis=1)

    # Compute confusion matrix
    cm = confusion_matrix(true_models, pred_models, normalize=normalize)

    # Initialize figure
    fig, ax = make_figure(1, 1, figsize=fig_size)
    ax = ax[0]
    im = ax.imshow(cm, interpolation="nearest", cmap=cmap)
    cbar = ax.figure.colorbar(im, ax=ax, shrink=0.75)

    cbar.ax.tick_params(labelsize=value_fontsize)

    ax.set_xticks(range(cm.shape[0]))
    ax.set_xticklabels(model_names, fontsize=tick_fontsize)
    if xtick_rotation:
        plt.xticks(rotation=xtick_rotation, ha="right")

    ax.set_yticks(range(cm.shape[1]))
    ax.set_yticklabels(model_names, fontsize=tick_fontsize)
    if ytick_rotation:
        plt.yticks(rotation=ytick_rotation)

    ax.set_xlabel("Predicted model", fontsize=label_fontsize)
    ax.set_ylabel("True model", fontsize=label_fontsize)

    # Loop over data dimensions and create text annotations
    fmt = ".2f" if normalize else "d"
    thresh = cm.max() / 2.0
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(
                j,
                i,
                format(cm[i, j], fmt),
                fontsize=value_fontsize,
                ha="center",
                va="center",
                color="white" if cm[i, j] > thresh else "black",
            )
    if title:
        ax.set_title("Confusion Matrix", fontsize=title_fontsize)
    return fig



import numpy as np
from scipy.stats import binom as scipy_binomial


def minimal_coverage_probs(z: np.ndarray, u: np.ndarray) -> np.ndarray:
    """Vectorized function to compute the minimal coverage probability for uniform
    empirical cumulative distribution functions (ECDFs) given evaluation points z
    and a sample of uniform draws.

    See: https://link.springer.com/article/10.1007/s11222-022-10090-6
    Säilynoja, T., Bürkner, P. C., & Vehtari, A. (2022).
    Graphical test for discrete uniformity and its applications in goodness-of-fit
    evaluation and multiple sample comparison. Statistics and Computing, 32(2), 32.

    Parameters
    ----------
    z  : np.ndarray of shape (num_points, )
        The vector of evaluation points.
    u  : np.ndarray of shape (num_simulations, num_samples)
        The matrix of simulated draws (samples) from U(0, 1)
    """

    N = u.shape[1]
    F_m = np.sum((z[:, np.newaxis] >= u[:, np.newaxis, :]), axis=-1) / u.shape[1]
    bin1 = scipy_binomial(N, z).cdf(N * F_m)
    bin2 = scipy_binomial(N, z).cdf(N * F_m - 1)
    gamma = 2 * np.min(np.min(np.stack([bin1, 1 - bin2], axis=-1), axis=-1), axis=-1)
    return gamma



import keras
from tests.utils import check_combination_simulator_adapter


def test_approximator_estimate(approximator, simulator, batch_size, adapter):
    check_combination_simulator_adapter(simulator, adapter)

    num_batches = 4
    data = simulator.sample((num_batches * batch_size,))

    batch = adapter(data)
    batch = keras.tree.map_structure(keras.ops.convert_to_tensor, batch)
    approximator.build_from_data(batch)

    estimates = approximator.estimate(data)

    assert isinstance(estimates, dict)
    print(keras.tree.map_structure(keras.ops.shape, estimates))



"""Activation functions linking network output to estimates with architecturally enforced properties."""

from .ordered import Ordered
from .ordered_quantiles import OrderedQuantiles
from .positive_definite import PositiveDefinite

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



import keras
import pickle


def test_dataset_is_picklable(dataset):
    pickled = pickle.loads(pickle.dumps(dataset))

    assert type(pickled) is type(dataset)

    samples = dataset[0]  # dict of {param_name: param_value}
    pickled_samples = pickled[0]
    assert isinstance(samples, dict)
    assert isinstance(pickled_samples, dict)

    assert list(samples.keys()) == list(pickled_samples.keys())

    for key in samples.keys():
        assert keras.ops.shape(samples[key]) == keras.ops.shape(pickled_samples[key])


def test_dataset_works_in_fit(model, dataset):
    print(next(iter(dataset[0].values())).dtype)
    model.fit(dataset, epochs=1, steps_per_epoch=1)


def test_dataset_returns_batch(dataset, batch_size):
    samples = dataset[0]  # dict of {param_name: param_value}
    samples = next(iter(samples.values()))  # first param value

    assert keras.ops.shape(samples)[0] == batch_size



import keras
import numpy as np

from bayesflow.adapters import Adapter
from bayesflow.simulators.simulator import Simulator


class OnlineDataset(keras.utils.PyDataset):
    """
    A dataset that is generated on-the-fly.
    """

    def __init__(
        self,
        simulator: Simulator,
        batch_size: int,
        num_batches: int,
        adapter: Adapter | None,
        *,
        stage: str = "training",
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.batch_size = batch_size
        self._num_batches = num_batches
        self.adapter = adapter
        self.simulator = simulator
        self.stage = stage

    def __getitem__(self, item: int) -> dict[str, np.ndarray]:
        batch = self.simulator.sample((self.batch_size,))

        if self.adapter is not None:
            batch = self.adapter(batch, stage=self.stage)

        return batch

    @property
    def num_batches(self) -> int:
        return self._num_batches



import pytest


def check_combination_simulator_adapter(simulator, adapter):
    """Make sure simulator and adapter fixtures fit together and appropriate errors are raised if not."""
    # check whether the simulator returns a 'weight' key
    simulator_with_sample_weight = "weight" in simulator.sample(1).keys()
    # scan adapter representation for occurance of a rename pattern for 'sample_weight'
    adapter_with_sample_weight = "-> 'sample_weight'" in str(adapter)

    if not simulator_with_sample_weight and adapter_with_sample_weight:
        # adapter should expect a 'weight' key and raise a KeyError.
        with pytest.raises(KeyError):
            adapter(simulator.sample(1))
        # Don't use this fixture combination for further tests.
        pytest.skip(reason="Do not use this fixture combination for further tests")  # TODO: better reason
    elif simulator_with_sample_weight and not adapter_with_sample_weight:
        # When a weight key is present, but the adapter does not configure it
        # to be used as sample weight, no error is raised currently.
        # Don't use this fixture combination for further tests.
        pytest.skip(reason="Do not use this fixture combination for further tests")  # TODO: better reason


def check_approximator_multivariate_normal_score(approximator):
    from bayesflow.approximators import PointApproximator
    from bayesflow.scores import MultivariateNormalScore

    if isinstance(approximator, PointApproximator):
        for score in approximator.inference_network.scores.values():
            if isinstance(score, MultivariateNormalScore):
                pytest.skip(reason="MultivariateNormalScore is unstable")



import re

import pytest
import io
from contextlib import redirect_stdout
from tests.utils import check_approximator_multivariate_normal_score


@pytest.mark.skip(reason="not implemented")
def test_compile(amortizer):
    amortizer.compile(optimizer="AdamW")


@pytest.mark.skip(reason="not implemented")
def test_fit(amortizer, dataset):
    amortizer.compile(optimizer="AdamW")
    amortizer.fit(dataset)

    assert amortizer.losses is not None


def test_loss_progress(approximator, train_dataset, validation_dataset):
    # as long as MultivariateNormalScore is unstable, skip fit progress test
    check_approximator_multivariate_normal_score(approximator)

    approximator.compile(optimizer="AdamW")
    num_epochs = 3

    # Capture ostream and train model
    with io.StringIO() as stream:
        with redirect_stdout(stream):
            approximator.fit(dataset=train_dataset, validation_data=validation_dataset, epochs=num_epochs)

        output = stream.getvalue()

    print(output)

    # check that there is a progress bar
    assert "━" in output, "no progress bar"

    # check that the loss is shown
    assert "loss" in output
    assert re.search(r"\bloss: \d+\.\d+", output) is not None, "training loss not correctly shown"

    # check that validation loss is shown
    assert "val_loss" in output
    assert re.search(r"\bval_loss: \d+\.\d+", output) is not None, "validation loss not correctly shown"

    # check that the shown loss is not nan or zero
    assert re.search(r"\bnan\b", output) is None, "found nan in output"
    assert re.search(r"\bloss: 0\.0000e\+00\b", output) is None, "found zero loss in output"



import pytest

import bayesflow as bf


def should_skip():
    import keras

    if keras.backend.backend() != "torch":
        return True, "Mamba tests can only be run on PyTorch."

    import torch

    if not torch.cuda.is_available():
        return True, "Mamba tests can only be run on GPU."

    try:
        import mamba_ssm  # noqa: F401
    except ImportError:
        return True, "Could not import mamba."

    return False, None


skip, reason = should_skip()


@pytest.mark.skipif(skip, reason=reason)
def test_mamba_summary(random_time_series, mamba_summary_network):
    out = mamba_summary_network(random_time_series)
    # Batch size 2, summary dim 4
    assert out.shape == (2, 4)


@pytest.mark.skipif(skip, reason=reason)
def test_mamba_trains(random_time_series, inference_network, mamba_summary_network):
    workflow = bf.BasicWorkflow(
        inference_network=inference_network,
        summary_network=mamba_summary_network,
        inference_variables=["parameters"],
        summary_variables=["observables"],
        simulator=bf.simulators.SIR(),
    )

    history = workflow.fit_online(epochs=2, batch_size=8, num_batches_per_epoch=2)
    assert "loss" in list(history.history.keys())
    assert len(history.history["loss"]) == 2



from collections.abc import Sequence, Mapping

import matplotlib.pyplot as plt
import numpy as np

from bayesflow.utils import prepare_plot_data, add_titles_and_labels, prettify_subplots


def z_score_contraction(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    figsize: Sequence[int] = None,
    label_fontsize: int = 16,
    title_fontsize: int = 18,
    tick_fontsize: int = 12,
    color: str = "#132a70",
    num_col: int = None,
    num_row: int = None,
) -> plt.Figure:
    """
    Implements a graphical check for global model sensitivity by plotting the
    posterior z-score over the posterior contraction for each set of posterior
    samples in ``estimates`` according to [1].

    - The definition of the posterior z-score is:

    post_z_score = (posterior_mean - true_parameters) / posterior_std

    And the score is adequate if it centers around zero and spreads roughly
    in the interval [-3, 3]

    - The definition of posterior contraction is:

    post_contraction = 1 - (posterior_variance / prior_variance)

    In other words, the posterior contraction is a proxy for the reduction in
    uncertainty gained by replacing the prior with the posterior.
    The ideal posterior contraction tends to 1.
    Contraction near zero indicates that the posterior variance is almost
    identical to the prior variance for the particular marginal parameter
    distribution.

    Note:
    Means and variances will be estimated via their sample-based estimators.

    [1] Schad, D. J., Betancourt, M., & Vasishth, S. (2021).
    Toward a principled Bayesian workflow in cognitive science.
    Psychological methods, 26(1), 103.

    Paper also available at https://arxiv.org/abs/1904.12765

    Parameters
    ----------
    estimates      : np.ndarray of shape (num_datasets, num_post_draws, num_params)
        The posterior draws obtained from num_datasets
    targets     : np.ndarray of shape (num_datasets, num_params)
        The prior draws (true parameters) used for generating the num_datasets
    variable_keys       : list or None, optional, default: None
       Select keys from the dictionaries provided in estimates and targets.
       By default, select all keys.
    variable_names    : list or None, optional, default: None
        The parameter names for nice plot titles. Inferred if None
    figsize           : tuple or None, optional, default : None
        The figure size passed to the matplotlib constructor. Inferred if None.
    label_fontsize    : int, optional, default: 16
        The font size of the y-label text
    title_fontsize    : int, optional, default: 18
        The font size of the title text
    tick_fontsize     : int, optional, default: 12
        The font size of the axis ticklabels
    color             : str, optional, default: '#8f2727'
        The color for the true vs. estimated scatter points and error bars
    num_row           : int, optional, default: None
        The number of rows for the subplots. Dynamically determined if None.
    num_col           : int, optional, default: None
        The number of columns for the subplots. Dynamically determined if None.

    Returns
    -------
    f : plt.Figure - the figure instance for optional saving

    Raises
    ------
    ShapeError
        If there is a deviation from the expected shapes of ``estimates`` and ``targets``.
    """

    # Gather plot data and metadata into a dictionary
    plot_data = prepare_plot_data(
        estimates=estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
        num_col=num_col,
        num_row=num_row,
        figsize=figsize,
    )

    estimates = plot_data.pop("estimates")
    targets = plot_data.pop("targets")

    # Estimate posterior means and stds
    post_means = estimates.mean(axis=1)
    post_vars = estimates.var(axis=1, ddof=1)
    post_stds = np.sqrt(post_vars)

    # Estimate prior variance
    prior_vars = targets.var(axis=0, keepdims=True, ddof=1)

    # Compute contraction and z-score
    contraction = np.clip(1 - (post_vars / prior_vars), 0, 1)
    z_score = (post_means - targets) / post_stds

    # Loop and plot
    for i, ax in enumerate(plot_data["axes"].flat):
        if i >= plot_data["num_variables"]:
            break

        ax.scatter(contraction[:, i], z_score[:, i], color=color, alpha=0.5)
        ax.set_xlim([-0.05, 1.05])

    prettify_subplots(plot_data["axes"], num_subplots=plot_data["num_variables"], tick_fontsize=tick_fontsize)

    # Add labels, titles, and set font sizes
    add_titles_and_labels(
        axes=plot_data["axes"],
        num_row=plot_data["num_row"],
        num_col=plot_data["num_col"],
        title=plot_data["variable_names"],
        xlabel="Posterior contraction",
        ylabel="Posterior z-score",
        title_fontsize=title_fontsize,
        label_fontsize=label_fontsize,
    )

    plot_data["fig"].tight_layout()
    return plot_data["fig"]



from collections.abc import Sequence

import keras

from bayesflow.networks.summary_network import SummaryNetwork
from bayesflow.types import Tensor
from bayesflow.utils.serialization import serializable

from .mamba_block import MambaBlock


@serializable
class Mamba(SummaryNetwork):
    """
    Wraps a sequence of Mamba modules using the simple Mamba module from:
    https://github.com/state-spaces/mamba/blob/main/mamba_ssm/modules/mamba_simple.py

    Copyright (c) 2023, Tri Dao, Albert Gu.

    Example usage in a BayesFlow workflow as a summary network:

    `summary_net = bayesflow.wrappers.Mamba(summary_dim=32)`
    """

    def __init__(
        self,
        summary_dim: int = 16,
        feature_dims: Sequence[int] = (64, 64),
        state_dims: Sequence[int] = (64, 64),
        conv_dims: Sequence[int] = (64, 64),
        expand_dims: Sequence[int] = (2, 2),
        bidirectional: bool = True,
        dt_min: float = 0.001,
        dt_max: float = 0.1,
        dropout: float = 0.05,
        device: str = "cuda",
        **kwargs,
    ):
        """
        A time-series summarization network using Mamba-based State Space Models (SSM). This model processes
        sequential input data using a sequence of Mamba SSM layers (determined by the length of the tuples),
        followed by optional pooling, dropout, and a dense layer for extracting summary statistics.

        Parameters
        ----------
        summary_dim : Sequence[int], optional
            The output dimensionality of the summary statistics layer (default is 16).
        feature_dims : Sequence[int], optional
            The feature dimension for each mamba block, default is (64, 64),
        state_dims : Sequence[int], optional
            The dimensionality of the internal state in each Mamba block, default is (64, 64)
        conv_dims : Sequence[int], optional
            The dimensionality of the convolutional layer in each Mamba block, default is (32, 32)
        expand_dims : Sequence[int], optional
            The expansion factors for the hidden state in each Mamba block, default is (2, 2)
        dt_min : float, optional
            Minimum dynamic state evolution over time (default is 0.001).
        dt_max : float, optional
            Maximum dynamic state evolution over time (default is 0.1).
        pooling : bool, optional
            Whether to apply global average pooling (default is True).
        dropout : int, float, or None, optional
            Dropout rate applied before the summary layer (default is 0.5).
        dropout: float, optional
            Dropout probability; dropout is applied to the pooled summary vector.
        device : str, optional
            The computing device. Currently, only "cuda" is supported (default is "cuda").
        **kwargs :
            Additional keyword arguments passed to the `SummaryNetwork` parent class.
        """

        super().__init__(**kwargs)

        if device != "cuda":
            raise NotImplementedError("MambaSSM only supports cuda as `device`.")

        self.mamba_blocks = []
        for feature_dim, state_dim, conv_dim, expand in zip(feature_dims, state_dims, conv_dims, expand_dims):
            mamba = MambaBlock(feature_dim, state_dim, conv_dim, expand, bidirectional, dt_min, dt_max, device)
            self.mamba_blocks.append(mamba)

        self.pooling_layer = keras.layers.GlobalAveragePooling1D()
        self.dropout = keras.layers.Dropout(dropout)
        self.summary_stats = keras.layers.Dense(summary_dim)

    def call(self, time_series: Tensor, training: bool = False, **kwargs) -> Tensor:
        """
        Apply a sequence of Mamba blocks, followed by pooling, dropout, and summary statistics.

        Parameters
        ----------
        time_series : Tensor
            Input tensor representing the time series data, typically of shape
            (batch_size, sequence_length, feature_dim).
        training : bool, optional
            Whether the model is in training mode (default is False). Affects the behavior of
            the inner dropout and norm layers.
        **kwargs : dict
            Additional keyword arguments (not used in this method).

        Returns
        -------
        Tensor
            Output tensor after applying Mamba blocks, pooling, dropout, and summary statistics.
        """

        summary = time_series
        for mamba_block in self.mamba_blocks:
            summary = mamba_block(summary, training=training)

        summary = self.pooling_layer(summary)
        summary = self.dropout(summary, training=training)
        summary = self.summary_stats(summary)

        return summary



import os

import keras

import bayesflow as bf


def test_basic_workflow(tmp_path, inference_network, summary_network):
    workflow = bf.BasicWorkflow(
        inference_network=inference_network,
        summary_network=summary_network,
        inference_variables=["parameters"],
        summary_variables=["observables"],
        simulator=bf.simulators.SIR(),
        checkpoint_filepath=str(tmp_path),
    )

    # Ensure metrics work fine
    history = workflow.fit_online(epochs=4, batch_size=8, num_batches_per_epoch=2, verbose=0)
    plots = workflow.plot_default_diagnostics(test_data=50, num_samples=25)
    metrics = workflow.compute_default_diagnostics(test_data=50, num_samples=25, variable_names=["p1", "p2"])

    assert "loss" in list(history.history.keys())
    assert len(history.history["loss"]) == 4
    assert list(plots.keys()) == ["losses", "recovery", "calibration_ecdf", "z_score_contraction"]
    assert list(metrics.columns) == ["p1", "p2"]
    assert metrics.values.shape == (3, 2)

    # Ensure saving and loading from workflow works fine
    loaded_approximator = keras.saving.load_model(os.path.join(str(tmp_path), "model.keras"))

    # Get samples
    samples = loaded_approximator.sample(conditions=workflow.simulate(5), num_samples=3)
    assert samples["parameters"].shape == (5, 3, 2)



import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform


@serializable
class Log(ElementwiseTransform):
    """Log transforms a variable.

    Parameters
    ----------
    p1 : boolean
        Add 1 to the input before taking the logarithm?

    Examples
    --------
    >>> adapter = bf.Adapter().log(["x"])
    """

    def __init__(self, *, p1: bool = False):
        super().__init__()
        self.p1 = p1

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        if self.p1:
            return np.log1p(data)
        else:
            return np.log(data)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        if self.p1:
            return np.expm1(data)
        else:
            return np.exp(data)

    def get_config(self) -> dict:
        return serialize({"p1": self.p1})



from keras import ops

from bayesflow.types import Shape, Tensor
from bayesflow.utils.serialization import serializable

from ..invertible_layer import InvertibleLayer


@serializable
class OrthogonalPermutation(InvertibleLayer):
    """Implements a learnable orthogonal transformation according to [1]. Can be
    used as an alternative to a fixed ``Permutation`` layer.

    [1] Kingma, D. P., & Dhariwal, P. (2018). Glow: Generative flow with invertible 1x1
    convolutions. Advances in neural information processing systems, 31.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.weight = None

    def build(self, xz_shape: Shape, **kwargs) -> None:
        self.weight = self.add_weight(shape=(xz_shape[-1], xz_shape[-1]), initializer="orthogonal", trainable=True)

    def call(self, xz: Tensor, inverse: bool = False, **kwargs) -> (Tensor, Tensor):
        if inverse:
            return self._inverse(xz)
        return self._forward(xz)

    def _forward(self, x: Tensor) -> (Tensor, Tensor):
        z = ops.matmul(x, self.weight)
        log_det = ops.log(ops.abs(ops.det(self.weight)))

        if ops.ndim(x) > 2:
            log_det = ops.multiply(log_det, ops.shape(x)[1])

        return z, log_det

    def _inverse(self, z: Tensor) -> (Tensor, Tensor):
        weight = ops.inv(self.weight)

        x = ops.matmul(z, weight)
        log_det = ops.log(ops.abs(ops.det(weight)))

        if ops.ndim(z) > 2:
            log_det = ops.multiply(log_det, ops.shape(z)[1])

        return x, log_det



import keras
import logging
import matplotlib
import pytest

BACKENDS = ["jax", "numpy", "tensorflow", "torch"]


def pytest_runtest_setup(item):
    """Skips backends by test markers. Unmarked tests are treated as backend-agnostic"""
    backend = keras.backend.backend()

    test_backends = [mark.name for mark in item.iter_markers() if mark.name in BACKENDS]

    if test_backends and backend not in test_backends:
        pytest.skip(f"Skipping backend '{backend}' for test {item}, which is registered for backends {test_backends}.")

    # always show full tracebacks
    keras.config.disable_traceback_filtering()

    if keras.backend.backend() == "jax":
        import jax

        jax.config.update("jax_traceback_filtering", "off")

    # use a non-GUI plotting backend for tests
    matplotlib.use("Agg")

    # set the logging level to debug for all tests
    logging.getLogger("bayesflow").setLevel(logging.DEBUG)


def pytest_runtest_teardown(item, nextitem):
    import matplotlib.pyplot as plt

    # close all plots at the end of each test
    plt.close("all")


def pytest_make_parametrize_id(config, val, argname):
    return f"{argname}={repr(val)}"


@pytest.fixture(params=[2], scope="session")
def batch_size(request):
    return request.param


@pytest.fixture(params=[None, 2, 3], scope="session")
def conditions_size(request):
    return request.param


@pytest.fixture(params=[1, 4], scope="session")
def summary_dim(request):
    return request.param


@pytest.fixture(params=["two_moons"], scope="session")
def dataset(request):
    return request.getfixturevalue(request.param)


@pytest.fixture(params=[2, 3], scope="session")
def feature_size(request):
    return request.param


@pytest.fixture(scope="session")
def random_conditions(batch_size, conditions_size):
    if conditions_size is None:
        return None

    return keras.random.normal((batch_size, conditions_size))


@pytest.fixture(scope="session")
def random_samples(batch_size, feature_size):
    return keras.random.normal((batch_size, feature_size))


@pytest.fixture(scope="function", autouse=True)
def random_seed():
    seed = 0
    keras.utils.set_random_seed(seed)
    return seed


@pytest.fixture(scope="session")
def random_set(batch_size, set_size, feature_size):
    return keras.random.normal((batch_size, set_size, feature_size))


@pytest.fixture(params=[2, 3], scope="session")
def set_size(request):
    return request.param



import keras

from bayesflow.utils import filter_kwargs


match keras.backend.backend():
    case "jax":
        from .jax_approximator import JAXApproximator as BaseBackendApproximator
    case "numpy":
        from .numpy_approximator import NumpyApproximator as BaseBackendApproximator
    case "tensorflow":
        from .tensorflow_approximator import TensorFlowApproximator as BaseBackendApproximator
    case "torch":
        from .torch_approximator import TorchApproximator as BaseBackendApproximator
    case other:
        raise ValueError(f"Backend '{other}' is not supported.")


class BackendApproximator(BaseBackendApproximator):
    # noinspection PyMethodOverriding
    def fit(self, *, dataset: keras.utils.PyDataset, **kwargs):
        return super().fit(x=dataset, y=None, **filter_kwargs(kwargs, super().fit))



import inspect
from collections.abc import Callable, Mapping, Sequence
from typing import TypeVar, Any

import keras
import numpy as np

from bayesflow.types import Tensor
from . import logging

T = TypeVar("T")


def convert_args(f: Callable, *args: any, **kwargs: any) -> tuple[any, ...]:
    """Convert positional and keyword arguments to just positional arguments for f"""
    if not kwargs:
        return args

    signature = inspect.signature(f)

    # convert to just kwargs first
    kwargs = convert_kwargs(f, *args, **kwargs)

    parameters = []
    for name, param in signature.parameters.items():
        if param.kind in [param.VAR_POSITIONAL, param.VAR_KEYWORD]:
            continue

        parameters.append(kwargs.get(name, param.default))

    return tuple(parameters)


def convert_kwargs(f: Callable, *args: any, **kwargs: any) -> dict[str, any]:
    """Convert positional and keyword arguments qto just keyword arguments for f"""
    if not args:
        return kwargs

    signature = inspect.signature(f)

    parameters = dict(zip(signature.parameters, args))

    for name, value in kwargs.items():
        if name in parameters:
            raise TypeError(f"{f.__name__}() got multiple arguments for argument '{name}'")

        parameters[name] = value

    return parameters


def filter_kwargs(kwargs: dict[str, T], f: Callable) -> dict[str, T]:
    """Filter keyword arguments for f"""
    signature = inspect.signature(f)

    for parameter in signature.parameters.values():
        if parameter.kind == inspect.Parameter.VAR_KEYWORD:
            # there is a **kwargs parameter, so anything is valid
            return kwargs

    kwargs = {key: value for key, value in kwargs.items() if key in signature.parameters}

    return kwargs


def layer_kwargs(kwargs: dict[str, T]) -> dict[str, T]:
    """Filter keyword arguments for keras.Layer"""
    valid_keys = ["dtype", "name", "trainable"]
    return {key: value for key, value in kwargs.items() if key in valid_keys}


def model_kwargs(kwargs: dict[str, T]) -> dict[str, T]:
    """Filter keyword arguments for keras.Model"""
    return layer_kwargs(kwargs)


def sequential_kwargs(kwargs: dict[str, T]) -> dict[str, T]:
    """Filter keyword arguments for keras.Sequential"""
    valid_keys = ["name", "trainable"]
    return {key: value for key, value in kwargs.items() if key in valid_keys}


# TODO: rename and streamline and make protected
def check_output(outputs: T) -> None:
    # Warn if any NaNs present in output
    for k, v in outputs.items():
        nan_mask = keras.ops.isnan(v)
        if keras.ops.any(nan_mask):
            logging.warning("Found a total of {n:d} nan values for output {k}.", n=int(keras.ops.sum(nan_mask)), k=k)

    # Warn if any inf present in output
    for k, v in outputs.items():
        inf_mask = keras.ops.isinf(v)
        if keras.ops.any(inf_mask):
            logging.warning("Found a total of {n:d} inf values for output {k}.", n=int(keras.ops.sum(inf_mask)), k=k)


def split_tensors(data: Mapping[any, Tensor], axis: int = -1) -> Mapping[any, Tensor]:
    """Split tensors in the dictionary along the given axis."""
    result = {}

    for key, value in data.items():
        if keras.ops.shape(value)[axis] == 1:
            result[key] = keras.ops.squeeze(value, axis=axis)
            continue

        splits = keras.ops.split(value, keras.ops.shape(value)[axis], axis=axis)
        splits = [keras.ops.squeeze(split, axis=axis) for split in splits]

        for i, split in enumerate(splits):
            result[f"{key}_{i + 1}"] = split

    return result


def split_arrays(data: Mapping[str, np.ndarray], axis: int = -1) -> Mapping[str, np.ndarray]:
    """Split tensors in the dictionary along the given axis."""
    result = {}

    for key, value in data.items():
        if not hasattr(value, "shape"):
            result[key] = np.array([value])
            continue

        if len(value.shape) == 1:
            result[key] = value
            continue

        if value.shape[axis] == 1:
            result[key] = np.squeeze(value, axis=axis)
            continue

        splits = np.split(value, value.shape[axis], axis=axis)
        splits = [np.squeeze(split, axis=axis) for split in splits]

        for i, split in enumerate(splits):
            result[f"{key}_{i}"] = split

    return result


class VariableArray(np.ndarray):
    """
    An enriched numpy array with information on variable keys and names
    to be used in post-processing, specifically the diagnostics module.

    The current implemention is very basic and we may want to extend it
    in the future should this general structure prove useful.

    Design according to
    https://numpy.org/doc/stable/user/basics.subclassing.html#simple-example-adding-an-extra-attribute-to-ndarray
    """

    def __new__(cls, input_array, variable_keys=None, variable_names=None):
        obj = np.asarray(input_array).view(cls)
        obj.variable_keys = variable_keys
        obj.variable_names = variable_names
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.variable_keys = getattr(obj, "variable_keys", None)
        self.variable_names = getattr(obj, "variable_names", None)


def make_variable_array(
    x: Mapping[str, np.ndarray] | np.ndarray,
    dataset_ids: Sequence[int] | int = None,
    variable_keys: Sequence[str] | str = None,
    variable_names: Sequence[str] | str = None,
    default_name: str = "v",
) -> VariableArray:
    """
    Helper function to validate arrays for use in the diagnostics module.

    Parameters
    ----------
    x   : dict[str, ndarray] or ndarray. Dict of arrays or array to be validated.
        See dicts_to_arrays
    dataset_ids : Sequence of integers indexing the datasets to select (default = None).
        By default, use all datasets.
    variable_names : Sequence[str], optional (default = None)
        Optional variable names to act as a filter if dicts provided or actual variable names in case of array
        inputs.
    default_name   : str, optional (default = "v")
        The default variable name to use if array arguments and no variable names are provided.
    """

    if isinstance(variable_keys, str):
        variable_keys = [variable_keys]

    if isinstance(variable_names, str):
        variable_names = [variable_names]

    if isinstance(x, dict):
        if variable_keys is not None:
            x = {k: x[k] for k in variable_keys}

        variable_keys = x.keys()

        if dataset_ids is not None:
            if isinstance(dataset_ids, int):
                # dataset_ids needs to be a sequence so that np.stack works correctly
                dataset_ids = [dataset_ids]

            x = {k: v[dataset_ids] for k, v in x.items()}

        x = split_arrays(x)

        if variable_names is None:
            variable_names = list(x.keys())

        x = np.stack(list(x.values()), axis=-1)

    # Case arrays provided
    elif isinstance(x, np.ndarray):
        if isinstance(x, VariableArray):
            # reuse existing variable keys and names if contained in x
            if variable_names is None:
                variable_names = x.variable_names
            if variable_keys in None:
                variable_keys = x.variable_keys

        # use default names if not otherwise specified
        if variable_names is None:
            variable_names = [f"{default_name}_{i}" for i in range(x.shape[-1])]

        if dataset_ids is not None:
            x = x[dataset_ids]

    # Throw if unknown type
    else:
        raise TypeError(f"Only dicts and tensors are supported as arguments, but your estimates are of type {type(x)}")

    if len(variable_names) != x.shape[-1]:
        raise ValueError("Length of 'variable_names' should be the same as the number of variables.")

    if variable_keys is None:
        # every variable will count as its own key if not otherwise specified
        variable_keys = variable_names

    x = VariableArray(x, variable_keys=variable_keys, variable_names=variable_names)

    return x


def dicts_to_arrays(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray = None,
    priors: Mapping[str, np.ndarray] | np.ndarray = None,
    dataset_ids: Sequence[int] | int = None,
    variable_keys: Sequence[str] | str = None,
    variable_names: Sequence[str] | str = None,
    default_name: str = "v",
) -> dict[str, Any]:
    """Helper function that prepares estimates and optional ground truths for diagnostics
    (plotting or computation of metrics).

    The function operates on both arrays and dictionaries and assumes either a dictionary
    where each key contains a 1D or a 2D array (i.e., a univariate quantity or samples thereof)
    or a 2D or 3D array where the last axis represents all quantities of interest.

    If a `ground_truths` array is provided, it must correspond to estimates in terms of type
    and structure of the first and last axis.

    If a dictionary is provided, `variable_names` acts as a filter to select variables from
    estimates. If an array is provided, `variable_names` can be used to override the `default_name`.

    Parameters
    ----------
    estimates   : dict[str, ndarray] or ndarray
        The model-generated predictions or estimates, which can take the following forms:
        - ndarray of shape (num_datasets, num_variables)
            Point estimates for each dataset, where `num_datasets` is the number of datasets
            and `num_variables` is the number of variables per dataset.
        - ndarray of shape (num_datasets, num_draws, num_variables)
            Posterior samples for each dataset, where `num_datasets` is the number of datasets,
            `num_draws` is the number of posterior draws, and `num_variables` is the number of variables.

    targets : dict[str, ndarray] or ndarray, optional (default = None)
        Ground-truth values corresponding to the estimates. Must match the structure and dimensionality
        of `estimates` in terms of first and last axis.

    dataset_ids : Sequence of integers indexing the datasets to select (default = None).
        By default, use all datasets.

    variable_keys : list or None, optional, default: None
       Select keys from the dictionary provided in samples.
       By default, select all keys.

    variable_names : Sequence[str], optional (default = None)
        Optional variable names to act as a filter if dicts provided or actual variable names in case of array
        inputs.

    default_name   : str, optional (default = "v")
        The default variable name to use if array arguments and no variable names are provided.
    """

    # other to be validated arrays (see below) will take use
    # the variable_keys and variable_names implied by estimates
    estimates = make_variable_array(
        estimates,
        dataset_ids=dataset_ids,
        variable_keys=variable_keys,
        variable_names=variable_names,
        default_name=default_name,
    )

    if targets is not None:
        targets = make_variable_array(
            targets,
            dataset_ids=dataset_ids,
            variable_keys=estimates.variable_keys,
            variable_names=estimates.variable_names,
        )

    if priors is not None:
        priors = make_variable_array(
            priors,
            # priors are data independent so datasets_ids is not passed here
            variable_keys=estimates.variable_keys,
            variable_names=estimates.variable_names,
        )

    return dict(
        estimates=estimates,
        targets=targets,
        priors=priors,
    )


def squeeze_inner_estimates_dict(estimates):
    """If a dictionary has only one key-value pair and the key is "value", return only its value.
    Otherwise, return the unchanged dictionary.

    This method helps to remove unnecessary nesting levels.
    """
    if len(estimates.keys()) == 1 and "value" in estimates.keys():
        return estimates["value"]
    else:
        return estimates



import keras
from functools import singledispatch


@singledispatch
def find_permutation(arg, *args, **kwargs):
    raise TypeError(f"Cannot infer permutation from {arg!r}.")


@find_permutation.register
def _(name: str, *args, **kwargs):
    match name.lower():
        case "random":
            from bayesflow.networks.coupling_flow.permutations import RandomPermutation

            return RandomPermutation(*args, **kwargs)
        case "swap":
            from bayesflow.networks.coupling_flow.permutations import Swap

            return Swap(*args, **kwargs)
        case "learnable" | "orthogonal":
            from bayesflow.networks.coupling_flow.permutations import OrthogonalPermutation

            return OrthogonalPermutation(*args, **kwargs)


@find_permutation.register
def _(permutation: keras.Layer, *args, **kwargs):
    return permutation


@find_permutation.register
def _(none: None, *args, **kwargs):
    return None



import keras
import numpy as np


def test_two_moons(two_moons_simulator, batch_size):
    samples = two_moons_simulator.sample((batch_size,))

    assert isinstance(samples, dict)
    assert list(samples.keys()) == ["parameters", "observables"]
    assert all(isinstance(value, np.ndarray) for value in samples.values())

    assert samples["parameters"].shape == (batch_size, 2)
    assert samples["observables"].shape == (batch_size, 2)


def test_gaussian_linear(gaussian_linear_simulator, batch_size):
    samples = gaussian_linear_simulator.sample((batch_size,))

    # test n_obs respected if applicable
    if hasattr(gaussian_linear_simulator, "n_obs") and isinstance(gaussian_linear_simulator.n_obs, int):
        assert samples["observables"].shape[1] == gaussian_linear_simulator.n_obs


def test_sample(simulator, batch_size):
    samples = simulator.sample((batch_size,))

    # test output structure
    assert isinstance(samples, dict)

    for key, value in samples.items():
        print(f"{key}.shape = {keras.ops.shape(value)}")

        # test type
        assert isinstance(value, np.ndarray)

        # test shape
        assert value.shape[0] == batch_size

        # test batch randomness
        assert not np.allclose(value, value[0])


def test_fixed_sample(composite_gaussian, batch_size, fixed_n, fixed_mu):
    samples = composite_gaussian.sample((batch_size,), n=fixed_n, mu=fixed_mu)

    assert samples["n"] == fixed_n
    assert samples["mu"].shape == (batch_size, 1)
    assert np.all(samples["mu"] == fixed_mu)
    assert samples["y"].shape == (batch_size, fixed_n)



import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from .elementwise_transform import ElementwiseTransform


@serializable
class ExpandDims(ElementwiseTransform):
    """
    Expand the shape of an array.

    Parameters
    ----------
    axis : int or tuple
        The axis to expand.

    Examples
    --------
    shape (3,) array:

    >>> a = np.array([1, 2, 3])

    shape (2, 3) array:

    >>> b = np.array([[1, 2, 3], [4, 5, 6]])
    >>> dat = dict(a=a, b=b)

    >>> ed = bf.adapters.transforms.ExpandDims("a", axis=0)
    >>> new_dat = ed.forward(dat)
    >>> new_dat["a"].shape
    (1, 3)

    >>> ed = bf.adapters.transforms.ExpandDims("a", axis=1)
    >>> new_dat = ed.forward(dat)
    >>> new_dat["a"].shape
    (3, 1)

    >>> ed = bf.adapters.transforms.ExpandDims("b", axis=1)
    >>> new_dat = ed.forward(dat)
    >>> new_dat["b"].shape
    (2, 1, 3)

    It is recommended to precede this transform with a :class:`bayesflow.adapters.transforms.ToArray` transform.
    """

    def __init__(self, *, axis: int | tuple):
        super().__init__()
        self.axis = axis

    def get_config(self) -> dict:
        return serialize({"axis": self.axis})

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return np.expand_dims(data, axis=self.axis)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return np.squeeze(data, axis=self.axis)



from collections.abc import Mapping, Sequence

import numpy as np
import matplotlib.pyplot as plt

from ...utils.plot_utils import prepare_plot_data, add_titles_and_labels, prettify_subplots
from ...utils.ecdf import pointwise_ecdf_bands


def calibration_ecdf_from_quantiles(
    estimates: Mapping[str, Mapping[str, np.ndarray]],
    targets: Mapping[str, np.ndarray],
    quantile_levels: Sequence[float],
    quantiles_key: str = "quantiles",
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    difference: bool = False,
    stacked: bool = False,
    figsize: Sequence[float] = None,
    label_fontsize: int = 16,
    legend_fontsize: int = 14,
    legend_location: str = "upper right",
    title_fontsize: int = 18,
    tick_fontsize: int = 12,
    rank_ecdf_color: str = "#132a70",
    fill_color: str = "grey",
    num_row: int = None,
    num_col: int = None,
    **kwargs,
) -> plt.Figure:
    """
    Creates the empirical CDFs for each marginal rank distribution
    and plots it against a uniform ECDF.

    For models with many parameters, use `stacked=True` to obtain an idea
    of the overall calibration of a posterior approximator.

    Note: In contrast to the related calibration_ecdf() function, this does not use
    simultaneous confidence bands. Confidence bands apply to each quantile level separately.

    [1] Säilynoja, T., Bürkner, P. C., & Vehtari, A. (2022). Graphical test
    for discrete uniformity and its applications in goodness-of-fit evaluation
    and multiple sample comparison. Statistics and Computing, 32(2), 1-21.
    https://arxiv.org/abs/2103.10522

    [2] Lemos, Pablo, et al. "Sampling-based accuracy testing of posterior estimators
     for general inference." International Conference on Machine Learning. PMLR, 2023.
     https://proceedings.mlr.press/v202/lemos23a.html

    Parameters
    ----------
    estimates         : dict[str, dict[str, np.ndarray]]
        The model-generated estimates in a nested dictionary, e.g. as returned
        by approximator.estimate(conditions=...).
        - The outer keys identify the inference variable.
        - The inner keys identify point estimates.
          Select which one to treat as quantile predictions with argument quantiles_key.
        - The inner value is an ndarray of shape (num_datasets, point_estimate_size, variable_block_size)
    targets           : dict[str, np.ndarray]
        The prior draws (true parameters) used for generating the num_datasets
    quantile_levels   : list of floats
        The target quantile levels that the quantile predictions should be tested against.
    quantiles_key     : str, optional, default: "quantiles"
          Selects which estimate to treat as quantile predictions with argument quantiles_key.
    variable_keys       : list or None, optional, default: None
       Select keys from the dictionaries provided in estimates and targets.
       By default, select all keys.
    variable_names    : list or None, optional, default: None
        The parameter names for nice plot titles.
        Inferred if None. Only relevant if `stacked=False`.
    difference        : bool, optional, default: False
        If `True`, plots the ECDF difference.
        Enables a more dynamic visualization range.
    stacked           : bool, optional, default: False
        If `True`, all ECDFs will be plotted on the same plot.
        If `False`, each ECDF will have its own subplot,
        similar to the behavior of `calibration_histogram`.
    figsize           : tuple or None, optional, default: None
        The figure size passed to the matplotlib constructor.
        Inferred if None.
    label_fontsize    : int, optional, default: 16
        The font size of the y-label and y-label texts
    legend_fontsize   : int, optional, default: 14
        The font size of the legend text
    title_fontsize    : int, optional, default: 18
        The font size of the title text.
        Only relevant if `stacked=False`
    tick_fontsize     : int, optional, default: 12
        The font size of the axis ticklabels
    rank_ecdf_color   : str, optional, default: '#a34f4f'
        The color to use for the rank ECDFs
    fill_color        : str, optional, default: 'grey'
        The color of the fill arguments.
    num_row           : int, optional, default: None
        The number of rows for the subplots.
        Dynamically determined if None.
    num_col           : int, optional, default: None
        The number of columns for the subplots.
        Dynamically determined if None.
    **kwargs          : dict, optional, default: {}
        Keyword arguments can be passed to control the behavior of
        ECDF simultaneous band computation through the ``ecdf_bands_kwargs``
        dictionary. See `pointwise_ecdf_bands` for keyword arguments.

    Returns
    -------
    f : plt.Figure - the figure instance for optional saving

    Raises
    ------
    ShapeError
        If there is a deviation form the expected shapes of `estimates`
        and `targets`.
    ValueError
        If an unknown `rank_type` is passed.
    """

    estimates = {k: v[quantiles_key] for k, v in estimates.items()}

    plot_data = prepare_plot_data(
        estimates=estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
        num_col=num_col,
        num_row=num_row,
        figsize=figsize,
        stacked=stacked,
    )

    estimates = plot_data.pop("estimates")
    targets = plot_data.pop("targets")

    # Plot individual ecdf of parameters
    for j in range(estimates.shape[-1]):
        xx = quantile_levels
        yy = np.mean(estimates[:, :, j] > targets[:, None, j], axis=0)

        # Difference, if specified
        if difference:
            yy -= xx

        if stacked:
            if j == 0:
                plot_data["axes"][0].plot(xx, yy, marker="o", color=rank_ecdf_color, alpha=0.95, label="Rank ECDFs")
            else:
                plot_data["axes"][0].plot(xx, yy, marker="o", color=rank_ecdf_color, alpha=0.95)
        else:
            plot_data["axes"].flat[j].plot(xx, yy, marker="o", color=rank_ecdf_color, alpha=0.95, label="Rank ECDF")

    # Compute uniform ECDF and bands
    alpha, z, L, U = pointwise_ecdf_bands(estimates.shape[0], **kwargs.pop("ecdf_bands_kwargs", {}))

    # Difference, if specified
    if difference:
        L -= z
        U -= z
        ylab = "ECDF Difference"
    else:
        ylab = "ECDF"

    # Add simultaneous bounds
    if not stacked:
        titles = plot_data["variable_names"]
    else:
        titles = ["Stacked ECDFs"]

    for ax, title in zip(plot_data["axes"].flat, titles):
        ax.fill_between(
            z,
            L,
            U,
            color=fill_color,
            alpha=0.2,
            label=rf"{int((1 - alpha) * 100)}$\%$ Confidence Bands" + "\n(pointwise)",
        )
        ax.legend(fontsize=legend_fontsize, loc=legend_location)
        ax.set_title(title, fontsize=title_fontsize)

    prettify_subplots(plot_data["axes"], num_subplots=plot_data["num_variables"], tick_fontsize=tick_fontsize)

    add_titles_and_labels(
        plot_data["axes"],
        plot_data["num_row"],
        plot_data["num_col"],
        xlabel="Quantile level",
        ylabel=ylab,
        label_fontsize=label_fontsize,
    )

    plot_data["fig"].tight_layout()
    return plot_data["fig"]



import keras


def detailed_loss_callback() -> keras.callbacks.Callback:
    """
    Provides detailed callback for loss trajectory for every training step.
    """
    loss_callback = keras.callbacks.LambdaCallback(
        on_train_batch_end=lambda batch, logs: logs["loss"], on_test_batch_end=lambda batch, logs: logs["val_loss"]
    )

    return loss_callback


def custom_loss_callback() -> keras.callbacks.Callback:
    """
    Provides customizable callback for loss trajectory
    """
    raise NotImplementedError



from collections.abc import Sequence

import keras
from keras import ops, layers

from bayesflow.types import Tensor
from bayesflow.utils import layer_kwargs
from bayesflow.utils.decorators import sanitize_input_shape
from bayesflow.utils.serialization import serializable

from ..mlp import MLP

from .invariant_layer import InvariantLayer


@serializable
class EquivariantLayer(keras.Layer):
    """Implements an equivariant module performing an equivariant transform.

    For details and justification, see:

    [1] Bloem-Reddy, B., & Teh, Y. W. (2020). Probabilistic Symmetries and Invariant Neural Networks.
    J. Mach. Learn. Res., 21, 90-1. https://www.jmlr.org/papers/volume21/19-322/19-322.pdf
    """

    def __init__(
        self,
        mlp_widths_equivariant: Sequence[int] = (128, 128),
        mlp_widths_invariant_inner: Sequence[int] = (128, 128),
        mlp_widths_invariant_outer: Sequence[int] = (128, 128),
        pooling: str = "mean",
        activation: str = "gelu",
        kernel_initializer: str = "he_normal",
        dropout: int | float | None = 0.05,
        layer_norm: bool = True,
        spectral_normalization: bool = False,
        **kwargs,
    ):
        """
        Initializes an equivariant module that combines equivariant transformations with nested invariant transforms
        to enable interactions between set members.

        This module applies an equivariant transformation to each set member, followed by an invariant transformation
        that aggregates information and injects it back into the set representation.

        The architecture consists of a fully connected residual block for equivariant processing and an invariant
        module to enhance expressiveness.

        The model supports different activation functions, dropout, layer normalization, and optional spectral
        normalization for stability.

        Parameters
        ----------
        mlp_widths_equivariant : Sequence[int], optional
            Widths of the MLP layers in the equivariant transformation applied to each set member.
            Default is (128, 128).
        mlp_widths_invariant_inner : Sequence[int], optional
            Widths of the inner MLP layers in the invariant module. Default is (128, 128).
        mlp_widths_invariant_outer : Sequence[int], optional
            Widths of the outer MLP layers in the invariant module. Default is (128, 128).
        pooling : str, optional
            Type of pooling operation used in the invariant module, such as "mean". Default is "mean".
        activation : str, optional
            Activation function applied in the MLP layers, such as "gelu". Default is "gelu".
        kernel_initializer : str, optional
            Initialization strategy for kernel weights, such as "he_normal". Default is "he_normal".
        dropout : int, float, or None, optional
            Dropout rate applied within the MLP layers. Default is 0.05.
        layer_norm : bool, optional
            Whether to apply layer normalization after transformations. Default is True.
        spectral_normalization : bool, optional
            Whether to apply spectral normalization to stabilize training. Default is False.
        """

        super().__init__(**layer_kwargs(kwargs))

        # Invariant module to increase expressiveness by concatenating outputs to each set member
        self.invariant_module = InvariantLayer(
            mlp_widths_inner=mlp_widths_invariant_inner,
            mlp_widths_outer=mlp_widths_invariant_outer,
            activation=activation,
            kernel_initializer=kernel_initializer,
            dropout=dropout,
            pooling=pooling,
            spectral_normalization=spectral_normalization,
        )

        # Fully connected net + residual connection for an equivariant transform applied to each set member
        self.input_projector = layers.Dense(mlp_widths_equivariant[-1])
        self.equivariant_fc = MLP(
            mlp_widths_equivariant,
            dropout=dropout,
            activation=activation,
            kernel_initializer=kernel_initializer,
            spectral_normalization=spectral_normalization,
        )

        self.layer_norm = layers.LayerNormalization() if layer_norm else None

    @sanitize_input_shape
    def build(self, input_shape):
        self.call(keras.ops.zeros(input_shape))

    def call(self, input_set: Tensor, training: bool = False, **kwargs) -> Tensor:
        """Performs the forward pass of a learnable equivariant transform.

        Parameters
        ----------
        input_set : Tensor
            The input tensor representing a set, with shape
            (batch_size, ..., set_size, input_dim).
        training : bool, optional
            A flag indicating whether the model is in training mode. Default is False.
        **kwargs : dict
            Additional keyword arguments for compatibility with other functions.

        Returns
        -------
        Tensor
            The transformed output tensor with the same shape as `input_set`, where
            each element is processed through the equivariant transformation.
        """

        input_set = self.input_projector(input_set)

        # Store shape of input_set, will be (batch_size, ..., set_size, some_dim)
        shape = ops.shape(input_set)

        # Example: Output dim is (batch_size, ..., set_size, representation_dim)
        invariant_summary = self.invariant_module(input_set, training=training)
        invariant_summary = ops.expand_dims(invariant_summary, axis=-2)
        tiler = [1] * len(shape)
        tiler[-2] = shape[-2]
        invariant_summary = ops.tile(invariant_summary, tiler)

        # Concatenate each input entry with the repeated invariant embedding
        output_set = ops.concatenate([input_set, invariant_summary], axis=-1)

        # Pass through final equivariant transform + residual
        output_set = input_set + self.equivariant_fc(output_set, training=training)
        if self.layer_norm is not None:
            output_set = self.layer_norm(output_set, training=training)

        return output_set



import keras
from keras.saving import register_keras_serializable as serializable

from bayesflow.utils import layer_kwargs, logging

from collections.abc import Sequence

from .ordered import Ordered


@serializable(package="links.ordered_quantiles")
class OrderedQuantiles(Ordered):
    """Activation function to link to monotonously increasing quantile estimates."""

    def __init__(self, q: Sequence[float] = None, axis: int = None, **kwargs):
        super().__init__(axis, None, **layer_kwargs(kwargs))
        self.q = q

        self.config = {
            "q": q,
            "axis": axis,
        }

    def get_config(self):
        base_config = super().get_config()
        return base_config | self.config

    def build(self, input_shape):
        if self.axis is None and 1 < len(input_shape) <= 3:
            self.axis = -2
        elif self.axis is None:
            raise AssertionError(
                f"Cannot resolve which axis should be ordered automatically from input shape {input_shape}."
            )

        num_quantile_levels = input_shape[self.axis]

        if self.q is None:
            # choose the middle of the specified axis as anchor index
            self.anchor_index = num_quantile_levels // 2
            logging.info(
                f"`OrderedQuantiles` was not provided with argument `q`. Using index {self.anchor_index} as anchor."
            )
        else:
            # choose quantile level closest to median as anchor index
            self.anchor_index = keras.ops.argmin(keras.ops.abs(keras.ops.convert_to_tensor(self.q) - 0.5))

            if len(self.q) != num_quantile_levels:
                raise RuntimeError(
                    f"Length of `q` does not coincide with input shape: len(q)={len(self.q)}, "
                    f"position {self.axis} of shape={input_shape}"
                )

        if self.anchor_index in [0, -1, num_quantile_levels - 1]:
            raise RuntimeError(
                f"The link function `OrderedQuantiles` expects at least 3 quantile levels, "
                f"but only {num_quantile_levels} were given."
            )

        super().build(input_shape)



import keras.ops as ops

from bayesflow.types import Tensor
from bayesflow.utils.keras_utils import shifted_softplus
from bayesflow.utils.serialization import serializable

from .transform import Transform


@serializable
class AffineTransform(Transform):
    def __init__(self, clamp: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.clamp = clamp

    @property
    def params_per_dim(self):
        return 2

    def split_parameters(self, parameters: Tensor) -> dict[str, Tensor]:
        scale, shift = ops.split(parameters, 2, axis=-1)

        return {"scale": scale, "shift": shift}

    def constrain_parameters(self, parameters: dict[str, Tensor]) -> dict[str, Tensor]:
        scale = parameters["scale"]

        # soft clamp
        if self.clamp:
            scale = ops.arcsinh(scale)

        # constrain to positive values
        scale = shifted_softplus(scale)

        parameters["scale"] = scale
        return parameters

    def _forward(self, x: Tensor, parameters: dict[str, Tensor] = None) -> (Tensor, Tensor):
        z = parameters["scale"] * x + parameters["shift"]
        log_det = ops.sum(ops.log(parameters["scale"]), axis=-1)

        return z, log_det

    def _inverse(self, z: Tensor, parameters: dict[str, Tensor] = None) -> (Tensor, Tensor):
        x = (z - parameters["shift"]) / parameters["scale"]
        log_det = -ops.sum(ops.log(parameters["scale"]), axis=-1)

        return x, log_det



import keras
from keras.saving import register_keras_serializable
import numpy as np
from bayesflow.networks.mlp import MLP

from bayesflow.types import Shape, Tensor
from bayesflow.utils import layer_kwargs


@register_keras_serializable(package="bayesflow.networks.cif")
class ConditionalGaussian(keras.Layer):
    """Implements a conditional gaussian distribution with neural networks for
    the means and standard deviations respectively. Bulit in reference to [1].

    [1] R. Cornish, A. Caterini, G. Deligiannidis, & A. Doucet (2021).
    Relaxing Bijectivity Constraints with Continuously Indexed Normalising
    Flows.
    arXiv:1909.13833.
    """

    def __init__(self, depth: int = 4, width: int = 128, activation: str = "swish", **kwargs):
        """Creates an instance of a `ConditionalGaussian` with configurable
        `MLP` networks for the means and standard deviations.

        Parameters
        ----------
        depth: int, optional, default: 4
            The number of MLP hidden layers (minimum: 1)
        width: int, optional, default: 128
            The dimensionality of the MLP hidden layers
        activation: str, optional, default: "swish"
            The MLP activation function
        """

        super().__init__(**layer_kwargs(kwargs))
        self.means = MLP([width] * depth, activation=activation)
        self.stds = MLP([width] * depth, activation=activation)
        self.output_projector = keras.layers.Dense(None)

    def build(self, input_shape: Shape) -> None:
        self.means.build(input_shape)
        self.stds.build(input_shape)
        self.output_projector.units = input_shape[-1]

    def _diagonal_gaussian_log_prob(self, conditions: Tensor, means: Tensor, stds: Tensor) -> Tensor:
        batch_size = keras.ops.shape(conditions)[0]

        if keras.ops.shape(means)[0] != batch_size or keras.ops.shape(stds)[0] != batch_size:
            raise ValueError("Means and stds must have the same batch size as conditions.")

        flat_conditions = keras.ops.reshape(conditions, (batch_size, -1))
        flat_means = keras.ops.reshape(means, (batch_size, -1))
        flat_stds = keras.ops.reshape(stds, (batch_size, -1))

        flat_variances = flat_stds**2

        dim = keras.ops.shape(flat_conditions)[1]

        const_term = -0.5 * dim * np.log(2 * np.pi)
        log_det_terms = -0.5 * keras.ops.sum(keras.ops.log(flat_variances), axis=1)
        product_terms = -0.5 * keras.ops.sum((flat_conditions - flat_means) ** 2 / flat_variances, axis=1)

        return const_term + log_det_terms + product_terms

    def log_prob(self, x: Tensor, conditions: Tensor) -> Tensor:
        means = self.output_projector(self.means(conditions))
        stds = keras.ops.exp(self.output_projector(self.stds(conditions)))
        return self._diagonal_gaussian_log_prob(x, means, stds)

    def sample(self, conditions: Tensor, log_prob: bool = False) -> Tensor | tuple[Tensor, Tensor]:
        means = self.output_projector(self.means(conditions))
        stds = keras.ops.exp(self.output_projector(self.stds(conditions)))

        # re-parametrize
        samples = stds * keras.random.normal(keras.ops.shape(conditions)) + means

        if log_prob:
            log_p = self._diagonal_gaussian_log_prob(samples, means, stds)
            return samples, log_p

        return samples



import keras
import numpy as np

from bayesflow.types import Tensor


def inverse_shifted_softplus(x: Tensor, shift: float = np.log(np.e - 1), beta: float = 1.0, threshold: float = 20.0):
    """Inverse of the shifted softplus function."""
    return inverse_softplus(x, beta=beta, threshold=threshold) - shift


def inverse_softplus(x: Tensor, beta: float = 1.0, threshold: float = 20.0) -> Tensor:
    """Numerically stabilized inverse softplus function."""
    return keras.ops.where(beta * x > threshold, x, keras.ops.log(keras.ops.expm1(beta * x)) / beta)


def shifted_softplus(x: Tensor, shift: float = np.log(np.e - 1)) -> Tensor:
    """Shifted version of the softplus function such that shifted_softplus(0) = 1"""
    return keras.ops.softplus(x + shift)



import numpy as np
import pytest


@pytest.fixture()
def adapter():
    from bayesflow.adapters import Adapter
    import keras

    @keras.saving.register_keras_serializable("custom")
    def serializable_fn(x):
        return x

    d = (
        Adapter()
        .to_array()
        .as_set(["s1", "s2"])
        .broadcast("t1", to="t2")
        .as_time_series(["t1", "t2"])
        .convert_dtype("float64", "float32", exclude="o1")
        .concatenate(["x1", "x2"], into="x")
        .concatenate(["y1", "y2"], into="y")
        .expand_dims(["z1"], axis=2)
        .log("p1")
        .constrain("p2", lower=0)
        .apply(include="p2", forward="exp", inverse="log")
        .apply(include="p2", forward="log1p")
        .apply_serializable(include="x", forward=serializable_fn, inverse=serializable_fn)
        .scale("x", by=[-1, 2])
        .shift("x", by=2)
        .split("key_to_split", into=["split_1", "split_2"])
        .standardize(exclude=["t1", "t2", "o1"])
        .drop("d1")
        .one_hot("o1", 10)
        .keep(["x", "y", "z1", "p1", "p2", "s1", "s2", "t1", "t2", "o1", "split_1", "split_2"])
        .rename("o1", "o2")
    )

    return d


@pytest.fixture()
def random_data():
    return {
        "x1": np.random.standard_normal(size=(32, 1)),
        "x2": np.random.standard_normal(size=(32, 1)),
        "y1": np.random.standard_normal(size=(32, 2)),
        "y2": np.random.standard_normal(size=(32, 2)),
        "z1": np.random.standard_normal(size=(32, 2)),
        "p1": np.random.lognormal(size=(32, 2)),
        "p2": np.random.lognormal(size=(32, 2)),
        "s1": np.random.standard_normal(size=(32, 3, 2)),
        "s2": np.random.standard_normal(size=(32, 3, 2)),
        "t1": np.zeros((3, 2)),
        "t2": np.ones((32, 3, 2)),
        "d1": np.random.standard_normal(size=(32, 2)),
        "d2": np.random.standard_normal(size=(32, 2)),
        "o1": np.random.randint(0, 9, size=(32, 2)),
        "key_to_split": np.random.standard_normal(size=(32, 10)),
    }



import os
import pickle


def deserialize_type(config: dict) -> type:
    module = __import__(config["module"], fromlist=[config["name"]])
    return getattr(module, config["name"])


def format_bytes(b: int, precision: int = 2, si: bool = False) -> str:
    """
    Format a number of bytes as a human-readable string in the format '{value} {prefix}{unit}'.

    :param b: The number of bytes to format.
    :param precision: The display precision.
    :param si: Determines whether to use SI decimal or binary prefixes.

    Examples:
        >>> format_bytes(1024)
        '1.00 KiB'
        >>> format_bytes(1024, si=True)
        '1.02 kB'
    """
    if si:
        div = 1000
        prefixes = ["", "k", "M", "G", "T", "P", "E", "Z", "Y", "R", "Q"]
    else:
        div = 1024
        prefixes = ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi", "Ri", "Qi"]

    value = float(b)
    idx = 0
    while value >= div and idx < len(prefixes) - 1:
        value /= div
        idx += 1

    prefix = prefixes[idx]

    return f"{value:.{precision}f} {prefix}B"


def parse_bytes(s: str) -> int:
    """
    Parse a string in the format '{value} {prefix}{unit}' and return the number of bytes,
    flooring to the nearest integer.

    The parsing is case-sensitive. E.g., uppercase 'K' will *not* be recognized as 'kilo',
    and lowercase 'b' will be recognized as 'bit' rather than 'byte'.

    Examples:
        >>> parse_bytes("8 GB")  # 8 Gigabyte
        8000000000
        >>> parse_bytes("32 kiB")  # 32 Kibibyte
        32768
        >>> parse_bytes("1 Tb")  # 1 Terrabit
        125000000000
        >>> parse_bytes("2.5 kB")  # 2.5 Kilobyte
        2500
        >>> parse_bytes("1e9 B")  # 10^9 Bytes
        1000000000
    """
    s = s.strip()
    if s.count(" ") != 1:
        raise ValueError(
            "Cannot parse bytes from string without exactly one space separator. "
            "Expected format: '{value} {prefix}{unit}'."
            "Example: '8 GiB'."
        )

    value, unit = s.split(" ")

    value = float(value)
    prefix, suffix = unit[:-1], unit[-1]

    if "i" not in prefix:
        # SI
        prefixes = ["", "k", "M", "G", "T", "P", "E", "Z", "Y", "R", "Q"]
        factor = 1000 ** prefixes.index(prefix)
    else:
        prefixes = ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi", "Ri", "Qi"]
        factor = 1024 ** prefixes.index(prefix)

    result = int(value * factor)

    if suffix == "b":
        result //= 8

    return result


def pickle_load(filepath: os.PathLike) -> any:
    """Small utility to safely load a pickle object."""

    with open(filepath, "rb+") as f:
        return pickle.load(f)


def serialize_type(obj: any) -> dict:
    if not isinstance(obj, type):
        obj = type(obj)

    return {"module": obj.__module__, "name": obj.__qualname__}



from collections.abc import Sequence

import keras

import warnings

from bayesflow.distributions import Distribution
from bayesflow.types import Shape, Tensor
from bayesflow.utils import (
    expand_right_as,
    find_network,
    integrate,
    jacobian_trace,
    layer_kwargs,
    optimal_transport,
    weighted_mean,
)
from bayesflow.utils.serialization import serialize, deserialize, serializable
from ..inference_network import InferenceNetwork


@serializable
class FlowMatching(InferenceNetwork):
    """Implements Optimal Transport Flow Matching, originally introduced as Rectified Flow, with ideas incorporated
    from [1-3].

    [1] Rectified Flow: arXiv:2209.03003
    [2] Flow Matching: arXiv:2210.02747
    [3] Optimal Transport Flow Matching: arXiv:2302.00482
    """

    MLP_DEFAULT_CONFIG = {
        "widths": (256, 256, 256, 256, 256),
        "activation": "mish",
        "kernel_initializer": "he_normal",
        "residual": True,
        "dropout": 0.05,
        "spectral_normalization": False,
    }

    OPTIMAL_TRANSPORT_DEFAULT_CONFIG = {
        "method": "log_sinkhorn",
        "regularization": 0.1,
        "max_steps": 100,
        "atol": 1e-5,
        "rtol": 1e-4,
    }

    INTEGRATE_DEFAULT_CONFIG = {
        "method": "euler",
        "steps": 100,
    }

    def __init__(
        self,
        subnet: str | keras.Layer = "mlp",
        base_distribution: str | Distribution = "normal",
        use_optimal_transport: bool = False,
        loss_fn: str | keras.Loss = "mse",
        integrate_kwargs: dict[str, any] = None,
        optimal_transport_kwargs: dict[str, any] = None,
        subnet_kwargs: dict[str, any] = None,
        **kwargs,
    ):
        """
        Initializes a flow-based model with configurable subnet architecture, loss function, and optional optimal
        transport integration.

        This model learns a transformation from a base distribution to a target distribution using a specified subnet
        type, which can be an MLP or a custom network. It supports flow matching with optional optimal transport for
        improved sample efficiency.

        The integration and transport steps can be customized with additional parameters available in the respective
        configuration dictionaries.

        Parameters
        ----------
        subnet : str or keras.Layer, optional
            The architecture used for the transformation network. Can be "mlp" or a custom
            callable network. Default is "mlp".
        base_distribution : str, optional
            The base probability distribution from which samples are drawn, such as "normal".
            Default is "normal".
        use_optimal_transport : bool, optional
            Whether to apply optimal transport for improved training stability. Default is False.
            Note: this will increase training time by approximately ~2.5 times, but may lead to faster inference.
        loss_fn : str, optional
            The loss function used for training, such as "mse". Default is "mse".
        integrate_kwargs : dict[str, any], optional
            Additional keyword arguments for the integration process. Default is None.
        optimal_transport_kwargs : dict[str, any], optional
            Additional keyword arguments for configuring optimal transport. Default is None.
        subnet_kwargs: dict[str, any], optional, deprecated
            Keyword arguments passed to the subnet constructor or used to update the default MLP settings.
        **kwargs
            Additional keyword arguments passed to the subnet and other components.
        """
        super().__init__(base_distribution, **kwargs)

        self.use_optimal_transport = use_optimal_transport

        self.integrate_kwargs = FlowMatching.INTEGRATE_DEFAULT_CONFIG | (integrate_kwargs or {})
        self.optimal_transport_kwargs = FlowMatching.OPTIMAL_TRANSPORT_DEFAULT_CONFIG | (optimal_transport_kwargs or {})

        self.loss_fn = keras.losses.get(loss_fn)

        self.seed_generator = keras.random.SeedGenerator()

        if subnet_kwargs:
            warnings.warn(
                "Using `subnet_kwargs` is deprecated."
                "Instead, instantiate the network yourself and pass the arguments directly.",
                DeprecationWarning,
            )

        subnet_kwargs = subnet_kwargs or {}
        if subnet == "mlp":
            subnet_kwargs = FlowMatching.MLP_DEFAULT_CONFIG | subnet_kwargs

        self.subnet = find_network(subnet, **subnet_kwargs)
        self.output_projector = keras.layers.Dense(units=None, bias_initializer="zeros", name="output_projector")

    def build(self, xz_shape: Shape, conditions_shape: Shape = None) -> None:
        if self.built:
            # building when the network is already built can cause issues with serialization
            # see https://github.com/keras-team/keras/issues/21147
            return

        self.base_distribution.build(xz_shape)

        self.output_projector.units = xz_shape[-1]

        # account for concatenating the time and conditions
        input_shape = list(xz_shape)
        input_shape[-1] += 1
        if conditions_shape is not None:
            input_shape[-1] += conditions_shape[-1]
        input_shape = tuple(input_shape)

        self.subnet.build(input_shape)
        input_shape = self.subnet.compute_output_shape(input_shape)
        self.output_projector.build(input_shape)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()
        base_config = layer_kwargs(base_config)

        config = {
            "subnet": self.subnet,
            "base_distribution": self.base_distribution,
            "use_optimal_transport": self.use_optimal_transport,
            "loss_fn": self.loss_fn,
            "integrate_kwargs": self.integrate_kwargs,
            "optimal_transport_kwargs": self.optimal_transport_kwargs,
            # we do not need to store subnet_kwargs
        }

        return base_config | serialize(config)

    def velocity(self, xz: Tensor, time: float | Tensor, conditions: Tensor = None, training: bool = False) -> Tensor:
        time = keras.ops.convert_to_tensor(time, dtype=keras.ops.dtype(xz))
        time = expand_right_as(time, xz)
        time = keras.ops.broadcast_to(time, keras.ops.shape(xz)[:-1] + (1,))

        if conditions is None:
            xtc = keras.ops.concatenate([xz, time], axis=-1)
        else:
            xtc = keras.ops.concatenate([xz, time, conditions], axis=-1)

        return self.output_projector(self.subnet(xtc, training=training), training=training)

    def _velocity_trace(
        self, xz: Tensor, time: Tensor, conditions: Tensor = None, max_steps: int = None, training: bool = False
    ) -> (Tensor, Tensor):
        def f(x):
            return self.velocity(x, time=time, conditions=conditions, training=training)

        v, trace = jacobian_trace(f, xz, max_steps=max_steps, seed=self.seed_generator, return_output=True)

        return v, keras.ops.expand_dims(trace, axis=-1)

    def _forward(
        self, x: Tensor, conditions: Tensor = None, density: bool = False, training: bool = False, **kwargs
    ) -> Tensor | tuple[Tensor, Tensor]:
        if density:

            def deltas(time, xz):
                v, trace = self._velocity_trace(xz, time=time, conditions=conditions, training=training)
                return {"xz": v, "trace": trace}

            state = {"xz": x, "trace": keras.ops.zeros(keras.ops.shape(x)[:-1] + (1,), dtype=keras.ops.dtype(x))}
            state = integrate(deltas, state, start_time=1.0, stop_time=0.0, **(self.integrate_kwargs | kwargs))

            z = state["xz"]
            log_density = self.base_distribution.log_prob(z) + keras.ops.squeeze(state["trace"], axis=-1)

            return z, log_density

        def deltas(time, xz):
            return {"xz": self.velocity(xz, time=time, conditions=conditions, training=training)}

        state = {"xz": x}
        state = integrate(deltas, state, start_time=1.0, stop_time=0.0, **(self.integrate_kwargs | kwargs))

        z = state["xz"]

        return z

    def _inverse(
        self, z: Tensor, conditions: Tensor = None, density: bool = False, training: bool = False, **kwargs
    ) -> Tensor | tuple[Tensor, Tensor]:
        if density:

            def deltas(time, xz):
                v, trace = self._velocity_trace(xz, time=time, conditions=conditions, training=training)
                return {"xz": v, "trace": trace}

            state = {"xz": z, "trace": keras.ops.zeros(keras.ops.shape(z)[:-1] + (1,), dtype=keras.ops.dtype(z))}
            state = integrate(deltas, state, start_time=0.0, stop_time=1.0, **(self.integrate_kwargs | kwargs))

            x = state["xz"]
            log_density = self.base_distribution.log_prob(z) - keras.ops.squeeze(state["trace"], axis=-1)

            return x, log_density

        def deltas(time, xz):
            return {"xz": self.velocity(xz, time=time, conditions=conditions, training=training)}

        state = {"xz": z}
        state = integrate(deltas, state, start_time=0.0, stop_time=1.0, **(self.integrate_kwargs | kwargs))

        x = state["xz"]

        return x

    def compute_metrics(
        self,
        x: Tensor | Sequence[Tensor, ...],
        conditions: Tensor = None,
        sample_weight: Tensor = None,
        stage: str = "training",
    ) -> dict[str, Tensor]:
        if isinstance(x, Sequence):
            # already pre-configured
            x0, x1, t, x, target_velocity = x
        else:
            # not pre-configured, resample
            x1 = x
            if not self.built:
                xz_shape = keras.ops.shape(x1)
                conditions_shape = None if conditions is None else keras.ops.shape(conditions)
                self.build(xz_shape, conditions_shape)
            x0 = self.base_distribution.sample(keras.ops.shape(x1)[:-1])

            if self.use_optimal_transport:
                # we must choose between resampling x0 or x1
                # since the data is possibly noisy and may contain outliers, it is better
                # to possibly drop some samples from x1 than from x0
                # in the marginal over multiple batches, this is not a problem
                x0, x1, assignments = optimal_transport(
                    x0,
                    x1,
                    seed=self.seed_generator,
                    **self.optimal_transport_kwargs,
                    return_assignments=True,
                )
                if conditions is not None:
                    # conditions must be resampled along with x1
                    conditions = keras.ops.take(conditions, assignments, axis=0)

            t = keras.random.uniform((keras.ops.shape(x0)[0],), seed=self.seed_generator)
            t = expand_right_as(t, x0)

            x = t * x1 + (1 - t) * x0
            target_velocity = x1 - x0

        base_metrics = super().compute_metrics(x1, conditions=conditions, stage=stage)

        predicted_velocity = self.velocity(x, time=t, conditions=conditions, training=stage == "training")

        loss = self.loss_fn(target_velocity, predicted_velocity)
        loss = weighted_mean(loss, sample_weight)

        return base_metrics | {"loss": loss}
